from pydantic import BaseModel
from alist_sdk.models import Storage, Setting, User, Meta


class Configs(BaseModel):
    settings: list[Setting] = []
    users: list[User] = []
    storages: list[Storage] = []
    metas: list[Meta] = []



from .client import Client
from .async_client import AsyncClient
from .models import *
from .err import *
from .version import __version__
from .path_lib import (
    AlistPath,
    PureAlistPath,
    AlistServer,
    login_server,
    AlistPathType,
    AbsAlistPathType,
)


__all__ = [
    "Client",
    "AsyncClient",
    "AlistPath",
    "PureAlistPath",
    "AlistServer",
    "login_server",
    "AlistPathType",
    "AbsAlistPathType",
    "__version__",
    *models.__all__,
    *err.__all__,
]



import asyncio
import logging
import time
import urllib.parse
from pathlib import Path, PurePosixPath

from httpx import AsyncClient as HttpClient, Response

from alist_sdk.models import *
from alist_sdk.verify import async_verify as verify
from alist_sdk.client import Client as SyncClient
from alist_sdk.version import __version__

logger = logging.getLogger("alist-sdk.async-client")

__all__ = ["AsyncClient"]


class AsyncClientBase(HttpClient):
    def __init__(
        self,
        base_url,
        token=None,
        username=None,
        password=None,
        has_opt=False,
        max_connect=30,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.base_url = base_url
        self.headers.setdefault("User-Agent", f"Alist-SDK/{__version__}")
        self.request_semaphore = asyncio.Semaphore(max_connect)
        if token or username:
            self.headers.update(
                SyncClient(
                    base_url=base_url,
                    token=token,
                    username=username,
                    password=password,
                    has_opt=has_opt,
                    **kwargs,
                ).headers
            )

    @staticmethod
    def asyncio_run(con):
        """添加事件循环"""
        try:
            asyncio.get_running_loop().run_in_executor(None, con)
        except RuntimeError:
            asyncio.run(con)

    async def verify_login_status(self) -> bool:
        """验证登陆状态,"""
        me = (await self.get("/api/me")).json()
        if me.get("code") != 200:
            logger.error(
                "异步客户端登陆失败[%d], %s", me.get("code"), me.get("message")
            )
            return False

        username = me["data"].get("username")
        if username not in [None]:
            logger.info("异步客户端登陆成功： 当前用户： %s", username)
            return True
        logger.warning("异步客户端登陆失败")
        return False

    async def set_token(self, token) -> bool:
        """更新Token，Token验证成功，返回True"""
        self.headers.update({"Authorization": token})
        return await self.verify_login_status()

    def get_token(self):
        return self.headers.get("Authorization")

    @property
    def server_info(self) -> tuple[str, str, int | None]:
        return self.base_url.scheme, self.base_url.host, self.base_url.port

    async def request(self, method: str, url, **kwargs) -> "Response":
        async with self.request_semaphore:
            return await super().request(method, url, **kwargs)

    @verify()
    async def verify_request(
        self,
        method: str,
        url,
        *,
        content=None,
        data=None,
        files=None,
        json=None,
        params=None,
        headers=None,
        follow_redirects=True,
        **kwargs,
    ):
        return {}, await self.request(
            method=method,
            url=url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            follow_redirects=follow_redirects,
            **kwargs,
        )

    @verify()
    async def me(self):
        return locals(), await self.get("/api/me")

    async def login(self, username, password, has_opt=False) -> bool:
        """登陆，成功返回Ture"""
        endpoint = "/api/auth/login"
        res = await self.post(
            url=endpoint,
            json={
                "username": username,
                "password": password,
                "opt_code": input("请输入当前OPT代码：") if has_opt else "",
            },
        )

        if res.status_code == 200 and res.json()["code"] == 200:
            return await self.set_token(res.json()["data"]["token"])
        logger.warning("登陆失败[%d]：%s", res.status_code, res.text)
        return False

    # ================ FS 相关方法 =================


class _AsyncFS(AsyncClientBase):
    @verify()
    async def mkdir(self, path: str | PurePosixPath):
        return locals(), await self.post("/api/fs/mkdir", json={"path": str(path)})

    @verify()
    async def rename(self, new_name, full_path: str | PurePosixPath):
        return locals(), await self.post(
            "/api/fs/rename",
            json={
                "name": new_name,
                "path": str(full_path),
            },
        )

    @verify()
    async def upload_file_form_data(
        self, data, path: str | PurePosixPath, as_task=False
    ):
        """表单上传"""
        return locals(), await self.put(
            "/api/fs/form",
            headers={
                "As-Task": "true" if as_task else "false",
                "File-Path": urllib.parse.quote_plus(str(path)),
            },
            # content=data,
            files={"upload-file": data},
        )

    @verify()
    async def upload_file_put(
        self, local_path: str | Path | bytes, path: str | PurePosixPath, as_task=False
    ):
        """流式上传文件"""
        if isinstance(local_path, bytes):
            data = local_path
            modified = int(time.time() * 1000)
        else:
            local_path = Path(local_path)
            if not local_path.exists():
                raise FileNotFoundError(local_path)
            data = local_path.read_bytes()
            modified = int(local_path.stat(follow_symlinks=True).st_mtime * 1000)

        return locals(), await self.put(
            "/api/fs/put",
            headers={
                "As-Task": "true" if as_task else "false",
                "Content-Type": "application/octet-stream",
                "Last-Modified": str(modified),
                "File-Path": urllib.parse.quote_plus(str(path)),
            },
            content=data,
        )

    @verify()
    async def list_files(
        self,
        path: str | PurePosixPath,
        password=None,
        page=1,
        per_page=0,
        refresh=False,
    ):
        """POST 列出文件目录"""
        return locals(), await self.post(
            "/api/fs/list",
            json={
                "path": str(path),
                "password": password,
                "page": page,
                "per_page": per_page,
                "refresh": refresh,
            },
        )

    @verify()
    async def get_item_info(self, path: str | PurePosixPath, password=None):
        """POST 获取某个文件/目录信息"""
        return locals(), await self.post(
            "/api/fs/get", json={"path": str(path), "password": password}
        )

    @verify()
    async def search(
        self,
        path: str | PurePosixPath,
        keyword: str,
        scope: SearchScopeModify = 0,
        page: int = None,
        per_page: int = None,
        password: str = None,
    ):
        """POST 搜索文件或文件夹

        :param password:
        :param per_page:
        :param page:
        :param keyword:
        :param path:
        :param scope: 搜索类型	0-全部 1-文件夹 2-文件
        """
        return locals(), await self.post(
            "/api/fs/search",
            json={
                "parent": str(path),
                "keywords": keyword,
                "scope": scope,
                "page": page,
                "per_page": per_page,
                "password": password,
            },
        )

    @verify()
    async def get_dir(
        self,
        path: str | PurePosixPath,
        password=None,
        page: int = 1,
        per_page: int = 10,
        refresh: bool = False,
    ):
        """POST 获取目录 /api/fs/dirs

        :return:
        """
        return locals(), await self.post(
            "/api/fs/dirs",
            json={
                "path": str(path),
                "password": password,
                "page": page,
                "per_page": per_page,
                "refresh": refresh,
            },
        )

    @verify()
    async def batch_rename(self):
        """POST 批量重命名  POST /api/fs/batch_rename
        https://alist.nn.ci/zh/guide/api/fs.html#post-%E6%89%B9%E9%87%8F%E9%87%8D%E5%91%BD%E5%90%8D
        """
        raise NotImplemented

    @verify()
    async def regex_rename(self):
        """POST 正则重命名   /api/fs/regex_rename"""
        raise NotImplemented

    @verify()
    async def move(
        self,
        src_dir: str | PurePosixPath,
        dst_dir: str | PurePosixPath,
        files: list[str] | str,
    ):
        """POST 移动文件  /api/fs/move"""
        return locals(), await self.post(
            "/api/fs/move",
            json={
                "src_dir": str(src_dir),
                "dst_dir": str(dst_dir),
                "names": [files] if isinstance(files, str) else files,
            },
        )

    @verify()
    async def recursive_move(
        self, src_dir: str | PurePosixPath, dst_dir: str | PurePosixPath
    ):
        """POST 聚合移动"""
        return locals(), await self.post(
            "/api/fs/recursive_move",
            json={
                "src_dir": str(src_dir),
                "dst_dir": str(dst_dir),
            },
        )

    @verify()
    async def copy(
        self,
        src_dir: str | PurePosixPath,
        dst_dir: str | PurePosixPath,
        files: list[str] | str,
    ):
        """POST 复制文件  /api/fs/copy"""

        return locals(), await self.post(
            "/api/fs/copy",
            json={
                "src_dir": str(src_dir),
                "dst_dir": str(dst_dir),
                "names": [files] if isinstance(files, str) else files,
            },
        )

    @verify()
    async def remove(self, path: str | PurePosixPath, names: list[str] | str):
        return locals(), await self.post(
            "/api/fs/remove",
            json={
                "names": [names] if isinstance(names, str) else names,
                "dir": str(path),
            },
        )

    @verify()
    async def remove_empty_directory(self, path: str | PurePosixPath):
        """/api/fs/remove_empty_directory"""
        return locals(), await self.post(
            "/api/fs/remove_empty_directory", json={"src_dir": str(path)}
        )

    @verify()
    async def add_aria2(self, path: str | PurePosixPath, urls: list[str] | str):
        return locals(), await self.post(
            "/api/fs/add_aria2",
            json={
                "urls": [urls] if isinstance(urls, str) else urls,
                "path": path,
            },
        )

    @verify()
    async def add_qbit(self, path: str | PurePosixPath, urls: list[str]):
        return locals(), await self.post(
            "/api/fs/add_qbit",
            json={
                "urls": [urls] if isinstance(urls, str) else urls,
                "path": str(path),
            },
        )

    # ================ admin/task 相关API ============================


class _AsyncAdminTask(AsyncClientBase):
    @staticmethod
    def task_type_verify(task_type: TaskTypeModify):
        if task_type in TaskTypeModify.__args__:  # type: ignore
            return True
        raise ValueError(f"{task_type = }, not in {TaskTypeModify}")

    @verify()
    async def task_done(self, task_type: TaskTypeModify):
        """获取已经完成的任务"""
        self.task_type_verify(task_type)
        return locals(), await self.get(f"/api/admin/task/{task_type}/done")

    @verify()
    async def task_undone(self, task_type: TaskTypeModify):
        """获取未完成的任务"""
        self.task_type_verify(task_type)
        return locals(), await self.get(f"/api/admin/task/{task_type}/undone")

    @verify()
    async def task_delete(self, task_type: TaskTypeModify, task_id):
        """删除任务"""
        self.task_type_verify(task_type)
        return locals(), await self.post(
            f"/api/admin/task/{task_type}/delete", params={"tid": task_id}
        )

    @verify()
    async def task_cancel(self, task_type: TaskTypeModify, task_id):
        """取消任务"""
        self.task_type_verify(task_type)
        return locals(), await self.post(
            f"/api/admin/task/{task_type}/cancel", params={"tid": task_id}
        )

    @verify()
    async def task_clear_done(self, task_type: TaskTypeModify):
        """清除已经完成的任务"""
        self.task_type_verify(task_type)
        return locals(), await self.post(f"/api/admin/task/{task_type}/clear_done")

    @verify()
    async def task_clear_succeeded(self, task_type: TaskTypeModify):
        """清除已成功的任务"""
        self.task_type_verify(task_type)
        return locals(), await self.post(f"/api/admin/task/{task_type}/clear_succeeded")

    @verify()
    async def task_retry(self, task_type: TaskTypeModify, task_id):
        """重试任务

        :return Resp -> data: task_id
        """
        self.task_type_verify(task_type)
        return locals(), await self.post(
            f"/api/admin/task/{task_type}/retry", params={"tid": task_id}
        )

    # ================= admin/storages 相关 ==========================


class _AsyncAdminStorage(AsyncClientBase):
    @verify()
    async def admin_storage_list(self):
        """列出存储器列表"""
        return locals(), await self.get("/api/admin/storage/list")

    @verify()
    async def admin_storage_create(self, storage: dict | Storage):
        """创建一个存储器后端"""
        if isinstance(storage, dict):
            storage = Storage(**storage)
        return locals(), await self.post(
            "/api/admin/storage/create",
            json=storage.model_dump(exclude={"id", "modified"}),
        )

    # ============== admin/user 相关==================


class _AsyncAdminUser(AsyncClientBase):
    @verify()
    async def admin_user_list(self):
        return locals(), await self.get("/api/admin/user/list")

    @verify()
    async def admin_user_add(self):
        """"""
        raise NotImplemented

    # ================== admin/meta 相关 ==============


class _AsyncAdminMeta(AsyncClientBase):
    @verify()
    async def admin_meta_list(self):
        return locals(), await self.get("/api/admin/meta/list")

    # ================== admin/setting 相关 =============


class _AsyncAdminSettings(AsyncClientBase):
    @verify()
    async def admin_setting_list(self, group: int = None):
        """"""

        query = {"group": group} if group else {}
        return locals(), await self.get("/api/admin/setting/list", params=query)


class AsyncClient(
    _AsyncFS,
    _AsyncAdminSettings,
    _AsyncAdminMeta,
    _AsyncAdminTask,
    _AsyncAdminStorage,
    _AsyncAdminUser,
):
    _cached_path_list = dict()
    _succeed_cache = 0

    async def dict_files_items(
        self,
        path: str | PurePosixPath,
        password="",
        refresh=False,
    ) -> dict[str, Item]:
        """列出文件目录"""
        path = str(path)
        if refresh:
            self._cached_path_list.pop(path, None)

        if path in self._cached_path_list:
            self._succeed_cache += 1
            logger.debug("缓存命中[times: %d]: %s", self._succeed_cache, path)
            return self._cached_path_list[path]

        if len(self._cached_path_list) >= 10000:
            self._cached_path_list.pop(0)  # Python 3中的字典是按照插入顺序保存的

        logger.debug("缓存未命中: %s", path)
        _res = await self.list_files(path, password, refresh=True)
        if _res.code == 200:
            _ = {d.name: d for d in _res.data.content or []}
            if _:  # 有数据才缓存
                self._cached_path_list[path] = _
            return _
        return {}



from alist_sdk import Client
from alist_sdk.tools.models import Configs
from alist_sdk.tools.configs import export_configs, import_configs_from_dict


class ExtraClient(Client):
    def export_configs(self) -> Configs:
        return export_configs(self)

    def export_config_json(self) -> str:
        return self.export_configs().model_dump_json(indent=2)

    def import_configs(self, configs):
        return import_configs_from_dict(self, configs)

    def import_config_from_other_client(
        self,
        base_url,
        token=None,
        username=None,
        password=None,
        has_opt=False,
        **kwargs,
    ):
        other_client = self.__class__(
            base_url, token, username, password, has_opt, **kwargs
        )
        return self.import_configs(
            other_client.export_configs().model_dump(mode="json")
        )



"""Path Lib 类实现

像使用Path一样的易于使用Alist中的文件
"""

import time
from functools import cached_property
from pathlib import Path
from typing import Iterator, Annotated, Any

from httpx import URL
from pydantic import BaseModel
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

from alist_sdk import alistpath, Item, AlistError, RawItem
from alist_sdk.py312_pathlib import PurePosixPath
from alist_sdk.client import Client


class AlistServer(BaseModel):
    server: str
    token: str
    kwargs: dict = {}


ALIST_SERVER_INFO: dict[tuple[str, str, int], Client] = dict()


def login_server(
    server: str | Client,
    token=None,
    username=None,
    password=None,
    has_opt=False,
    **kwargs,
) -> Client:
    """"""

    if isinstance(server, str):
        _so = URL(server)
        server_info = _so.scheme, _so.host, _so.port
        if server_info in ALIST_SERVER_INFO:
            return ALIST_SERVER_INFO[server_info]

        _client = Client(
            server,
            token=token,
            username=username,
            password=password,
            has_opt=has_opt,
            **kwargs,
        )

    else:
        _client = server
    ALIST_SERVER_INFO[_client.server_info] = _client
    return _client


class PureAlistPath(PurePosixPath):
    _flavour = alistpath

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.as_uri())

    def is_absolute(self):
        """True if the path is absolute (has both a root and, if applicable,
        a drive)."""
        if self._flavour is alistpath:
            return bool(self.drive and self.root)

        else:
            return self._flavour.isabs(str(self))

    def as_posix(self) -> str:
        return str(self).replace(self.drive, "")

    def as_uri(self):
        if not self.is_absolute():
            raise ValueError("relative path can't be expressed as a file URI")
        return str(self)

    def relative_to(self, other, /, *_deprecated, walk_up=False) -> str:
        other = self.with_segments(other, *_deprecated)
        for step, path in enumerate([other] + list(other.parents)):
            if self.is_relative_to(path):
                break

            elif path.name == "..":
                raise ValueError(f"'..' segment in {str(other)!r} cannot be walked")

        else:
            raise ValueError(f"{str(self)!r} and {str(other)!r} have different anchors")
        # noinspection PyProtectedMember
        parts = [".."] * step + self._tail[len(path._tail) :]
        return PurePosixPath(*parts).as_posix()

    # def match(self, path_pattern, *, case_sensitive=True):
    #     _pps = path_pattern.split("**")
    #
    #     return self._flavour.fnmatch(str(self), path_pattern, case_sensitive)


class AlistPath(PureAlistPath):
    """"""

    def __init__(
        self,
        *args,
        username: str = None,
        password: str = None,
        token: str = None,
        **kwargs,
    ):
        super().__init__(*args)
        if username:
            login_server(
                self.drive, username=username, password=password, token=token, **kwargs
            )

    @classmethod
    def from_client(cls, client: Client, path: str | PurePosixPath) -> "AlistPath":
        """从客户端实例构造

        :param client: 客户端实例
        :param path: （可能）位于该实例的绝对路径
        """
        if not isinstance(client, Client):
            raise TypeError()

        if not str(path).startswith("/"):
            raise ValueError(f"path必须是一个绝对路径 {path = }")

        if client.server_info not in ALIST_SERVER_INFO:
            login_server(client)

        f_path = client.base_url.join(path)
        return cls(f_path.__str__())

    @cached_property
    def client(self) -> Client:
        if self.drive == "":
            raise AlistError("当前对象没有设置server")

        try:
            _u = URL(self.drive)
            _server_info = _u.scheme, _u.host, _u.port
            return ALIST_SERVER_INFO[_server_info]
        except KeyError:
            raise AlistError(
                f"当前服务器[{self.drive}]尚未登陆。\n"
                f"使用AlistPath.from_client构造。\n"
                f"或构造AlistPath时传入username等相关信息。"
            )

    # def is_absolute(self) -> bool:
    def get_download_uri(self) -> str:
        if not self.is_absolute():
            raise ValueError("relative path can't be expressed as a file URI")
        if self.is_dir():
            raise IsADirectoryError()
        return self.raw_stat().raw_url

    def as_download_uri(self):
        return self.get_download_uri()

    def raw_stat(self, retry=1, timeout=0.1) -> RawItem:
        try:
            _raw = self.client.get_item_info(self.as_posix())
            if _raw.code == 200:
                data = _raw.data
                self.set_stat(data)
                return data
            if _raw.code == 500 and (
                "object not found" in _raw.message
                or "storage not found" in _raw.message
            ):
                raise FileNotFoundError(_raw.message)
            raise AlistError(_raw.message)
        except FileNotFoundError as _e:
            if retry > 0:
                time.sleep(timeout)
                return self.raw_stat(retry - 1)
            raise _e

    def stat(self) -> Item | RawItem:
        def f_stat() -> Item | RawItem:
            _r = self.client.dict_files_items(self.parent.as_posix()).get(self.name)
            if not _r:
                raise FileNotFoundError(f"文件不存在: {self.as_posix()} ")
            self.set_stat(_r)
            return _r

        _stat = getattr(self, "_stat", None)
        if isinstance(_stat, Item | RawItem):
            return _stat

        return f_stat()

    def set_stat(self, value: RawItem | Item):
        # noinspection PyAttributeOutsideInit
        self._stat = value

    def re_stat(self, retry=2, timeout=1) -> Item:
        if hasattr(self, "_stat"):
            delattr(self, "_stat")
        self.client.list_files(self.parent.as_posix(), per_page=1, refresh=True)
        return self.raw_stat(retry=retry, timeout=timeout)

    def is_dir(self):
        """"""
        return self.stat().is_dir

    def is_file(self):
        """"""
        return not self.stat().is_dir

    def is_link(self):
        raise NotImplementedError("AlistPath不支持连接.")

    def exists(self):
        """"""
        try:
            return bool(self.re_stat())
        except FileNotFoundError:
            return False

    def iterdir(self) -> Iterator["AlistPath"]:
        """"""
        if not self.is_dir():
            raise

        for item in self.client.dict_files_items(
            self.as_posix(), refresh=True
        ).values():
            _ = self.joinpath(item.name)
            _.set_stat(item)
            yield _

    def read_text(self):
        """"""
        return self.client.get(
            self.as_download_uri(),
            follow_redirects=True,
            headers={"authorization": ""},
        ).text

    def read_bytes(self):
        """"""
        return self.client.get(
            self.as_download_uri(),
            follow_redirects=True,
            headers={"authorization": ""},
        ).content

    def write_text(self, data: str, as_task=False):
        """"""
        return self.write_bytes(data.encode(), as_task=as_task)

    def write_bytes(self, data: bytes | Path, as_task=False):
        """"""

        _res = self.client.upload_file_put(data, self.as_posix(), as_task=as_task)
        if _res.code == 200:
            return self.re_stat()
        raise AlistError()

    def mkdir(self, parents=False, exist_ok=False):
        """"""
        if self.exists():
            if exist_ok:
                return
            raise FileExistsError(f"相同名称已存在: {self.as_posix()}")

        if parents is False and not self.parent.exists():
            raise FileNotFoundError()

        return self.client.mkdir(self.as_posix())

    def touch(self, exist_ok=True):
        """"""
        if not exist_ok and self.exists():
            raise FileExistsError(f"文件已存在: {self.as_posix()}")
        return self.write_bytes(b"", as_task=False)

    def unlink(self, missing_ok=False):
        """"""
        if not self.exists():
            if missing_ok:
                return
            raise FileNotFoundError(f"文件不存在: {self.as_posix()}")
        _data = self.client.remove(self.parent.as_posix(), self.name)
        if _data.code != 200:
            raise AlistError(_data.message)

    def rmdir(self, missing_ok=False):
        """"""
        if not self.exists():
            if missing_ok:
                return
            raise FileNotFoundError(f"文件不存在: {self.as_posix()}")
        _data = self.client.remove_empty_directory(self.as_posix())
        if _data.code != 200:
            raise AlistError(_data.message)

    def rename(self, target: "AlistPath"):
        """"""
        if self == target:
            return
        if not self.exists():
            raise FileNotFoundError(f"文件不存在: {self.as_uri()}")

        if self.parent != target.parent:
            _data = self.client.move(
                self.parent.as_posix(),
                target.parent.as_posix(),
                self.name,
            )
            if _data.code != 200:
                raise AlistError(_data.message)

        if self.name != target.name:
            _data = self.client.rename(
                target.name,
                target.parent.joinpath(self.name).as_posix(),
            )
            if _data.code != 200:
                raise AlistError(_data.message)

        return target


class AlistPathPydanticAnnotation:
    @classmethod
    def validate_alist_path(cls, v: Any, handler) -> AlistPath:
        if isinstance(v, AlistPath):
            return v
        s = handler(v)
        return AlistPath(s)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type, _handler
    ) -> core_schema.CoreSchema:
        assert source_type is AlistPath
        return core_schema.no_info_wrap_validator_function(
            cls.validate_alist_path,
            core_schema.str_schema(),
            serialization=core_schema.to_string_ser_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema, handler) -> JsonSchemaValue:
        return handler(core_schema.str_schema())


class AbsAlistPathPydanticAnnotation(AlistPathPydanticAnnotation):
    @classmethod
    def validate_alist_path(cls, v: Any, handler) -> AlistPath:
        _path = super().validate_alist_path(v, handler)
        if _path.drive:
            return _path
        else:
            raise ValueError("Invalid AlistPath")


AlistPathType = Annotated[AlistPath, AlistPathPydanticAnnotation]

AbsAlistPathType = Annotated[AlistPath, AbsAlistPathPydanticAnnotation]



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : alistpath.py
@Author     : LeeCQ
@Date-Time  : 2024/2/25 12:40

"""
import os
import posixpath
from urllib.parse import unquote

from posixpath import *

__all__ = [*posixpath.__all__, "splitroot"]


def splitroot(p):
    """Split a pathname into drive, root and tail. On Posix, drive is always
    empty; the root may be empty, a single slash, or two slashes. The tail
    contains anything after the root. For example:

        splitroot('foo/bar') == ('', '', 'foo/bar')
        splitroot('/foo/bar') == ('', '/', 'foo/bar')
        splitroot('//foo/bar') == ('', '//', 'foo/bar')
        splitroot('///foo/bar') == ('', '/', '//foo/bar')
        splitroot('http://server/path/to/file') == ('http://server', '/', 'path/to/file')
    """
    p = os.fspath(p)
    if isinstance(p, bytes):
        _sep = b"/"
        empty = b""
    else:
        _sep = "/"
        empty = ""

    if p.startswith("http://") or p.startswith("https://"):
        # Absolute path, e.g.: 'http://server/path/to/file'
        if p.count("/") < 3:
            p += "/"
        p = unquote(p)
        sps = p.split("/", 3)
        return "/".join(sps[:3]), "/", (sps[-1])

    elif p[:1] != _sep:
        # Relative path, e.g.: 'foo'
        return empty, empty, p

    elif p[1:2] != _sep or p[2:3] == _sep:
        # Absolute path, e.g.: '/foo', '///foo', '////foo', etc.
        return empty, _sep, p[1:]

    else:
        # Precisely two leading slashes, e.g.: '//foo'. Implementation defined per POSIX, see
        # https://pubs.opengroup.org/onlinepubs/9699919799/basedefs/V1_chap04.html#tag_04_13
        return empty, p[:2], p[2:]



# 测试 path_lib.py
import pytest

from alist_sdk.path_lib import PureAlistPath, AlistPath, login_server
from tests.test_client import DATA_DIR


class TestPureAlistPath:
    def test_pure_alist_path(self):
        path = PureAlistPath("https://server:5245/path/to/file")
        # 属性
        assert path.parts == ("https://server:5245/", "path", "to", "file")
        assert path.drive == "https://server:5245"
        assert path.root == "/"
        assert path.anchor == "https://server:5245/"
        assert path.parent == PureAlistPath("https://server:5245/path/to")
        assert path.name == "file"
        assert path.suffix == ""
        assert path.suffixes == []
        assert path.stem == "file"

    def test_pure_alist_path_method(self):
        path = PureAlistPath("https://server/path/to/file")

        assert path.as_posix() == "/path/to/file"
        assert path.as_uri() == "https://server/path/to/file"

        assert path.joinpath("another") == PureAlistPath(
            "https://server/path/to/file/another"
        )
        assert path.joinpath("another", "file", "path") == PureAlistPath(
            "https://server/path/to/file/another/file/path"
        )
        assert path.joinpath("another/file/path") == PureAlistPath(
            "https://server/path/to/file/another/file/path"
        )

    def test_relative_to(self):
        path = PureAlistPath("https://server/path/to/file")
        p = PureAlistPath("https://server/path")

        assert path.relative_to(p) == ("to/file")


class TestAlistPath:
    def setup_class(self):
        self.client = login_server(
            "http://localhost:5245",
            username="admin",
            password="123456",
        )
        DATA_DIR.joinpath("test.txt").write_text("123")

    def setup_method(self):
        self.client._cached_path_list = {}

    def test_read_text(self):
        DATA_DIR.joinpath("test.txt").write_text("123")
        path = AlistPath("http://localhost:5245/local/test.txt")
        assert path.read_text() == "123"

    def test_alist_path(self):
        path = AlistPath("http://localhost:5245/local/test.txt")
        assert path.read_bytes() == b"123"

    def test_write_text(self):
        path = AlistPath("http://localhost:5245/local/test_write_text.txt")
        path.write_text("123")
        assert DATA_DIR.joinpath("test_write_text.txt").read_text() == "123"

    def test_write_bytes(self):
        path = AlistPath("http://localhost:5245/local/test_write_bytes.txt")
        path.write_bytes(b"123")
        assert DATA_DIR.joinpath("test_write_bytes.txt").read_bytes() == b"123"

    def test_mkdir(self):
        path = AlistPath("http://localhost:5245/local/test_mkdir")
        path.mkdir()
        assert DATA_DIR.joinpath("test_mkdir").is_dir()

    def test_touch(self):
        path = AlistPath("http://localhost:5245/local/test_touch.txt")
        path.touch()
        assert DATA_DIR.joinpath("test_touch.txt").exists()

    def test_unlink(self):
        DATA_DIR.joinpath("test_unlink.txt").write_text("123")
        path = AlistPath("http://localhost:5245/local/test_unlink.txt")
        path.unlink()
        assert not DATA_DIR.joinpath("test_unlink.txt").exists()

    @pytest.mark.skip("Alist 接口不生效")
    def test_rmdir(self):
        DATA_DIR.joinpath("test_rmdir").mkdir()
        path = AlistPath("http://localhost:5245/local/test_rmdir")
        path.rmdir()
        assert not DATA_DIR.joinpath("test_rmdir").exists()

    def test_rename(self):
        DATA_DIR.joinpath("test_rename.txt").write_text("123")
        path = AlistPath("http://localhost:5245/local/test_rename.txt")
        path.rename(AlistPath("http://localhost:5245/local/test_rename_new.txt"))
        assert not DATA_DIR.joinpath("test_rename.txt").exists()
        assert DATA_DIR.joinpath("test_rename_new.txt").read_text() == "123"

    def test_re_stat(self):
        DATA_DIR.joinpath("test_re_stat.txt").write_text("123")
        path = AlistPath("http://localhost:5245/local/test_re_stat.txt")
        assert path.stat().size == 3
        DATA_DIR.joinpath("test_re_stat.txt").write_text("1234")
        assert path.re_stat().size == 4

    def test_from_client(self):
        DATA_DIR.joinpath("test_from_client.txt").write_text("123")
        path = AlistPath.from_client(self.client, "/local/test_from_client.txt")
        assert path.exists()
        assert path.client is self.client


def test_pydantic():
    from pydantic import BaseModel
    from alist_sdk.path_lib import AlistPathType

    class T(BaseModel):
        p: AlistPathType

    T(p="https://leecq.cn")
    T(p="ss/c")


def test_abs_path():
    from pydantic import BaseModel
    from alist_sdk.path_lib import AbsAlistPathType

    class T(BaseModel):
        p: AbsAlistPathType

    T(p="https://leecq.cn")


@pytest.mark.xfail()
def test_abs_path_err():
    from pydantic import BaseModel
    from alist_sdk.path_lib import AbsAlistPathType

    class T(BaseModel):
        p: AbsAlistPathType

    T(p="ss/c")



import asyncio
import time
from pathlib import Path, PurePosixPath
from functools import cached_property

import pytest

from alist_sdk import Client, AsyncClient
from alist_sdk import models

WORKDIR = Path(__file__).parent
DATA_DIR = WORKDIR.joinpath("alist/test_dir")
DATA_DIR_DST = WORKDIR.joinpath("alist/test_dir_dst")
DATA_DIR_DST.mkdir(exist_ok=True)
DATA_DIR.mkdir(exist_ok=True)


def create_storage_local(client_, mount_name, local_path: Path):
    local_storage = {
        "mount_path": f"/{mount_name}",
        "order": 0,
        "driver": "Local",
        "cache_expiration": 0,
        "status": "work",
        "addition": '{"root_folder_path":"%s","thumbnail":false,'
        '"thumb_cache_folder":"","show_hidden":true,'
        '"mkdir_perm":"750"}' % local_path.absolute().as_posix(),
        "remark": "",
        "modified": "2023-11-20T15:00:31.608106706Z",
        "disabled": False,
        "enable_sign": False,
        "order_by": "name",
        "order_direction": "asc",
        "extract_folder": "front",
        "web_proxy": False,
        "webdav_policy": "native_proxy",
        "down_proxy_url": "",
    }
    if f"/{mount_name}" not in [
        i["mount_path"]
        for i in client_.get("/api/admin/storage/list").json()["data"]["content"]
    ]:
        assert (
            client_.post("/api/admin/storage/create", json=local_storage)
            .json()
            .get("code")
            == 200
        ), "创建Storage失败。"
    else:
        print("已经创建，跳过 ...")


def clear_dir(path: Path):
    if not path.exists():
        return
    if path.is_dir():
        for item in path.iterdir():
            clear_dir(item)
        path.rmdir()
    else:
        try:
            path.unlink(missing_ok=True)
        except PermissionError:
            time.sleep(1)
            clear_dir(path)
        except NotADirectoryError | FileNotFoundError:
            pass


def setup_module() -> None:
    # os.system(f"bash {Path(__file__).parent.joinpath('init_alist.sh')}")
    clear_dir(DATA_DIR)
    clear_dir(DATA_DIR_DST)
    DATA_DIR.mkdir(exist_ok=True)
    DATA_DIR_DST.mkdir(exist_ok=True)

    _client = Client("http://localhost:5245", username="admin", password="123456")
    create_storage_local(_client, "local", DATA_DIR)
    create_storage_local(_client, "local_dst", DATA_DIR_DST)


# noinspection PyMethodMayBeStatic
class TestSyncClient:
    @cached_property
    def client(self) -> Client:
        return Client("http://localhost:5245", username="admin", password="123456")

    def setup_method(self) -> None:
        print("--- setup_method ---")
        clear_dir(DATA_DIR)
        clear_dir(DATA_DIR_DST)
        DATA_DIR.mkdir(exist_ok=True)
        DATA_DIR_DST.mkdir(exist_ok=True)

    def run(self, func, *args, **kwargs):
        print("run", func.__name__, end=": ")
        res = func(*args, **kwargs)
        if isinstance(res, models.Resp):
            print("RESP: ", res.model_dump_json())
        else:
            print(res)
        return res

    def test_login(self):
        _client = Client("http://localhost:5245", verify=False)
        assert _client.login(
            username="admin",
            password="123456",
        ), "登陆失败"

    def test_me(self):
        res = self.run(self.client.me)
        assert res.code == 200
        assert isinstance(res.data, models.Me), "数据结构错误。"

    def test_mkdir(self):
        res = self.run(self.client.mkdir, "/local/test_mkdir")
        assert res.code == 200
        assert DATA_DIR.joinpath("test_mkdir").exists(), "创建失败"

    def test_rename(self):
        source = DATA_DIR.joinpath("test_rename-s")
        dest = DATA_DIR.joinpath("test_rename-d")
        source.mkdir(exist_ok=True)
        clear_dir(dest)
        res = self.run(self.client.rename, "test_rename-d", "/local/test_rename-s")
        assert res.code == 200
        assert dest.exists()
        assert not source.exists()

    @pytest.mark.skip
    def test_upload_file_from_data(self):
        data = b"test_data"
        alist_full_name = "/local/upload_file_form_data"
        path_file = DATA_DIR.joinpath("upload_file_form_data")
        res = self.run(
            self.client.upload_file_form_data, data, alist_full_name, as_task=False
        )
        assert res.data is None
        assert res.code == 200
        assert path_file.exists()

    def test_upload_put(self):
        alist_full_name = "/local/upload_file_form_data"
        alist_data_path = DATA_DIR.joinpath("upload_file_form_data")

        res = self.run(
            self.client.upload_file_put, __file__, alist_full_name, as_task=False
        )
        assert res.code == 200
        assert alist_data_path.exists()

    def test_upload_put_as_task(self):
        alist_full_name = "/local/upload_file_form_data_task"
        alist_data_path = DATA_DIR.joinpath("upload_file_form_data_task")

        res = self.run(
            self.client.upload_file_put, __file__, alist_full_name, as_task=True
        )
        assert res.code == 200
        time.sleep(2)
        assert alist_data_path.exists()

    def test_list_dir(self):
        DATA_DIR.joinpath("test_list_dir_dir").mkdir()
        DATA_DIR.joinpath("test_list_dir_file").write_text("test")
        res = self.run(
            self.client.list_files,
            "/local",
        )
        assert res.code == 200
        assert isinstance(res.data, models.ListItem)
        assert isinstance(res.data.content[0], models.Item)

    def test_dict_list_dir(self):
        DATA_DIR.joinpath("test_list_dir_dir").mkdir()
        DATA_DIR.joinpath("test_list_dir_file").write_text("test")
        res = self.run(
            self.client.dict_files_items,
            "/local",
        )
        assert isinstance(res, dict)
        assert len(res) == 2

    def test_list_dir_null(self):
        DATA_DIR.joinpath("test_list_dir_null").mkdir()
        res = self.run(
            self.client.list_files,
            "/local/test_list_dir_null",
        )
        assert res.code == 200
        assert isinstance(res.data, models.ListItem)
        assert res.data.total == 0

    def test_get_item_info_dir(self):
        DATA_DIR.joinpath("test_ite_info_dir").mkdir()
        res = self.run(self.client.get_item_info, "/local/test_ite_info_dir")
        assert res.code == 200
        assert isinstance(res.data, models.RawItem)

    def test_get_item_info_file(self):
        DATA_DIR.joinpath("test_ite_info_file.txt").write_text("abc")
        res = self.run(self.client.get_item_info, "/local/test_ite_info_file.txt")
        assert res.code == 200
        assert isinstance(res, models.Resp)
        assert isinstance(res.data, models.RawItem)

    def test_get_dir(self):
        DATA_DIR.joinpath("test_ite_info_dir").mkdir()

        res = self.run(self.client.get_dir, "/local")
        assert res.code == 200
        assert isinstance(res.data, list)
        assert isinstance(res.data[0], models.DirItem)

    def test_move(self):
        DATA_DIR.joinpath("test_move_src").mkdir()
        DATA_DIR.joinpath("test_move_src/test.txt").write_text("abc")
        DATA_DIR.joinpath("test_move_dst").mkdir()

        res = self.run(
            self.client.move,
            src_dir="/local/test_move_src",
            dst_dir="/local/test_move_dst",
            files=["test.txt"],
        )

        assert res.code == 200
        assert DATA_DIR.joinpath("test_move_dst/test.txt").exists()
        assert not DATA_DIR.joinpath("test_move_src/test.txt").exists()

    def test_recursive_move(self):
        DATA_DIR.joinpath("test_move_src").mkdir()
        DATA_DIR.joinpath("test_move_src/test1.txt").write_text("abc")
        DATA_DIR.joinpath("test_move_src/test2.txt").write_text("abc")
        DATA_DIR.joinpath("test_move_src/test3.txt").write_text("abc")
        DATA_DIR.joinpath("test_move_dst").mkdir()

        res = self.run(
            self.client.recursive_move,
            src_dir="/local/test_move_src",
            dst_dir="/local/test_move_dst",
        )

        assert res.code == 200
        assert DATA_DIR.joinpath("test_move_dst/test1.txt").exists()
        assert not DATA_DIR.joinpath("test_move_src/test1.txt").exists()

    def test_copy(self):
        DATA_DIR.joinpath("test_copy_src").mkdir()
        DATA_DIR.joinpath("test_copy_src/test.txt").write_text("abc")
        DATA_DIR.joinpath("test_copy_dst").mkdir()

        res = self.run(
            self.client.copy,
            src_dir="/local/test_copy_src",
            dst_dir="/local/test_copy_dst",
            files=["test.txt"],
        )

        assert res.code == 200
        assert DATA_DIR.joinpath("test_copy_dst/test.txt").exists()
        assert DATA_DIR.joinpath("test_copy_src/test.txt").exists()

    def test_copy_path(self):
        DATA_DIR.joinpath("test_copy_src").mkdir()
        DATA_DIR.joinpath("test_copy_src/test.txt").write_text("abc")
        DATA_DIR.joinpath("test_copy_dst").mkdir()

        res = self.run(
            self.client.copy,
            src_dir=PurePosixPath("/local/test_copy_src"),
            dst_dir=PurePosixPath("/local/test_copy_dst"),
            files="test.txt",
        )

        assert res.code == 200
        assert DATA_DIR.joinpath("test_copy_dst/test.txt").exists()
        assert DATA_DIR.joinpath("test_copy_src/test.txt").exists()

    def test_remove_file(self):
        DATA_DIR.joinpath("test_remove_file.txt").write_text("abc")
        res = self.run(
            self.client.remove, path="/local", names=["test_remove_file.txt"]
        )
        assert res.code == 200
        assert not DATA_DIR.joinpath("test_remove_file.txt").exists()

    def test_remove_dir_has_file(self):
        DATA_DIR.joinpath("test_dir").mkdir()
        DATA_DIR.joinpath("test_dir/test1.txt").write_text("123")

        res = self.run(self.client.remove, path="/local", names=["test_dir"])
        assert res.code == 200
        assert not DATA_DIR.joinpath("test_dir/test1.txt").exists()
        assert not DATA_DIR.joinpath("test_dir").exists()

    @pytest.mark.skip
    def test_remove_empty_directory(self):
        DATA_DIR.joinpath("test_dir").mkdir()
        res = self.run(self.client.remove_empty_directory, path="/local/test_dir/")

        assert res.code == 200
        assert not DATA_DIR.joinpath("test_dir").exists()

    @pytest.mark.skip
    def test_remove_empty_directory_not(self):
        DATA_DIR.joinpath("test_dir").mkdir()
        DATA_DIR.joinpath("test_dir/test.txt").write_text("123")

        res = self.run(self.client.remove_empty_directory, path="/local/test_dir/")

        assert res.code == 200
        assert not DATA_DIR.joinpath("test_dir").exists()

    @pytest.mark.skip
    def test_add_aria2(self):
        res = self.run(self.client.add_aria2, "/local", [""])
        assert res.code == 200

    @pytest.mark.skip
    def test_add_qbit(self):
        """"""
        res = self.run(self.client.add_qbit, "/local", [""])
        assert res.code == 200

    @pytest.mark.parametrize(
        "task_type",
        models.TaskType,
    )
    def test_task_done(self, task_type):
        """"""
        DATA_DIR.joinpath("test.txt").write_text("abc")

        res = self.run(self.client.task_done, task_type)
        assert res.code == 200, res.message
        assert isinstance(res.data, list)
        # assert isinstance(res.data[0], models.Task)

    def test_storage_list(self):
        res = self.run(self.client.admin_storage_list)
        assert res.code == 200
        assert isinstance(res.data, models.ListContents)
        assert isinstance(res.data.content, (list, models.NoneType))
        assert isinstance(res.data.content[0], models.Storage)

        assert "/local" in [s.mount_path for s in res.data.content]

    def test_storage_create(self):
        """测试创建存储器"""
        import time

        store_name = f"test_storage_{int(time.time_ns())}"
        store_path = WORKDIR.joinpath(f"alist/{store_name}").absolute()
        store_path.mkdir(parents=True, exist_ok=True)
        local_storage = {
            "mount_path": f"/{store_name}",
            "order": 0,
            "driver": "Local",
            "cache_expiration": 0,
            "status": "work",
            "addition": '{"root_folder_path":"%s","thumbnail":false,'
            '"thumb_cache_folder":"","show_hidden":true,'
            '"mkdir_perm":"750"}' % store_path.as_posix(),
            "remark": "",
            "modified": "2023-11-20T15:00:31.608106706Z",
            "disabled": False,
            "enable_sign": False,
            "order_by": "name",
            "order_direction": "asc",
            "extract_folder": "front",
            "web_proxy": False,
            "webdav_policy": "native_proxy",
            "down_proxy_url": "",
        }
        res = self.run(self.client.admin_storage_create, local_storage)
        assert res.code == 200, res.message

    def test_admin_setting_list(self):
        res = self.run(self.client.admin_setting_list)
        assert res.code == 200

    def test_admin_user_list(self):
        res = self.run(self.client.admin_user_list)
        assert res.code == 200

    def test_admin_meta_list(self):
        res = self.run(self.client.admin_meta_list)
        assert res.code == 200


class TestAsyncClient(TestSyncClient):
    @property
    def client(self) -> AsyncClient:
        return AsyncClient("http://localhost:5245", username="admin", password="123456")

    def run(self, func, *args, **kwargs):
        print("--- type", type(func))
        return asyncio.run(func(*args, **kwargs))

    def test_async_client(self):
        assert isinstance(self.client, AsyncClient)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : path_lib_old.py
@Author     : LeeCQ
@Date-Time  : 2024/2/25 12:58

path_lib的存档文件。

Path Lib 类实现

像使用Path一样的易于使用Alist中的文件
"""
import fnmatch
import os
import posixpath
import re
import sys
from pathlib import PurePosixPath, PurePath
from functools import lru_cache, cached_property
from urllib.parse import quote_from_bytes as urlquote_from_bytes
from typing import Iterator

from pydantic import BaseModel
from httpx import URL

from alist_sdk import AlistError, RawItem
from alist_sdk.client import Client


class AlistServer(BaseModel):
    server: str
    token: str
    kwargs: dict = {}


ALIST_SERVER_INFO: dict[str, AlistServer] = dict()


def login_server(
        server: str,
        token=None,
        username=None,
        password=None,
        has_opt=False,
        **kwargs,
):
    """"""
    if token is None:
        client = Client(
            server,
            username=username,
            password=password,
            has_opt=has_opt,
            **kwargs,
        )
        token = client.headers.get("Authorization")
    ALIST_SERVER_INFO[server] = AlistServer(
        server=server,
        token=token,
        kwargs=kwargs,
    )


# noinspection PyMethodMayBeStatic
class _AlistFlavour:
    sep = "/"
    altsep = ""
    has_drv = True
    pathmod = posixpath

    is_supported = True

    def __init__(self):
        self.join = self.sep.join

    def parse_parts(self, parts):
        parsed = []
        sep = self.sep
        altsep = self.altsep
        drv = root = ""

        it = reversed(parts)
        for part in it:
            if not part:
                continue
            if altsep:
                part = part.replace(altsep, sep)
            drv, root, rel = self.splitroot(part)
            if sep in rel:
                for x in reversed(rel.split(sep)):
                    if x and x != ".":
                        parsed.append(sys.intern(x))
            else:
                if rel and rel != ".":
                    parsed.append(sys.intern(rel))
            if drv or root:
                if not drv:
                    # If no drive is present, try to find one in the previous
                    # parts. This makes the result of parsing e.g.
                    # ("C:", "/", "a") reasonably intuitive.
                    # noinspection PyAssignmentToLoopOrWithParameter
                    for part in it:
                        if not part:
                            continue
                        if altsep:
                            part = part.replace(altsep, sep)
                        drv = self.splitroot(part)[0]
                        if drv:
                            break
                break
        if drv or root:
            parsed.append(drv + root)
        parsed.reverse()
        return drv, root, parsed

    def join_parsed_parts(self, drv, root, parts, drv2, root2, parts2):
        """
        Join the two paths represented by the respective
        (drive, root, parts) tuples.  Return a new (drive, root, parts) tuple.
        """
        if root2:
            if not drv2 and drv:
                return drv, root2, [drv + root2] + parts2[1:]
        elif drv2:
            if drv2 == drv or self.casefold(drv2) == self.casefold(drv):
                # Same drive => second path is relative to the first
                return drv, root, parts + parts2[1:]
        else:
            # Second path is non-anchored (common case)
            return drv, root, parts + parts2
        return drv2, root2, parts2

    def splitroot(self, part, sep=sep):
        if part and part[0] == sep:
            stripped_part = part.lstrip(sep)
            # According to POSIX path resolution:
            # http://pubs.opengroup.org/onlinepubs/009695399/basedefs/xbd_chap04.html#tag_04_11
            # "A pathname that begins with two successive slashes may be
            # interpreted in an implementation-defined manner, although more
            # than two leading slashes shall be treated as a single slash".
            if len(part) - len(stripped_part) == 2:
                return "", sep * 2, stripped_part
            else:
                return "", sep, stripped_part
        else:
            return "", "", part

    def casefold(self, s):
        return s

    def casefold_parts(self, parts):
        return parts

    def compile_pattern(self, pattern):
        return re.compile(fnmatch.translate(pattern)).fullmatch

    def make_uri(self, path):
        # We represent the path using the local filesystem encoding,
        # for portability to other applications.
        bpath = bytes(path)
        return "file://" + urlquote_from_bytes(bpath)


# noinspection PyProtectedMember,PyUnresolvedReferences
class PureAlistPath(PurePosixPath):
    _flavour = _AlistFlavour()

    @classmethod
    def _parse_args(cls, args):
        # This is useful when you don't want to create an instance, just
        # canonicalize some constructor arguments.
        parts = []
        for a in args:
            if isinstance(a, PurePath):
                parts += a._parts
            else:
                a = os.fspath(a)
                if isinstance(a, str):
                    # Force-cast str subclasses to str (issue #21127)
                    parts.append(str(a))
                else:
                    raise TypeError(
                        "argument should be a str object or an os.PathLike "
                        "object returning str, not %r" % type(a)
                    )
        return cls._flavour.parse_parts(parts)

    @classmethod
    def _from_parts(cls, args):
        # We need to call _parse_args on the instance, to get the
        # right flavour.
        args = list(args)
        self = object.__new__(cls)
        if isinstance(args[0], str) and args[0].startswith("http"):
            _u = URL(args[0])
            server = (
                f"{_u.scheme}://{_u.host}:{_u.port}".replace(":80", "")
                .replace(":443", "")
                .replace(":None", "")
            )
            args[0] = _u.path
        elif isinstance(args[0], cls):
            server = args[0].server
        else:
            server = ""

        drv, root, parts = self._parse_args(args)
        self._drv = drv or server
        self._root = root
        self._parts = parts
        self.server = server
        return self

    @classmethod
    def _from_parsed_parts(cls, drv, root, parts):
        self = object.__new__(cls)
        self._drv = drv
        self._root = root
        self._parts = parts
        return self

    def _make_child(self, args):
        drv, root, parts = self._parse_args(args)
        drv, root, parts = self._flavour.join_parsed_parts(
            self._drv, self._root, self._parts, drv, root, parts
        )
        return self._from_parsed_parts(drv, root, parts)

    def __new__(cls, *args):
        return cls._from_parts(args)

    def as_posix(self) -> str:
        return str(self).replace(self.drive, "")

    def as_uri(self):
        if not self.is_absolute():
            raise ValueError("relative path can't be expressed as a file URI")
        return str(self)

    def relative_to(self, *other):
        raise NotImplementedError("AlistPath不支持relative_to")


class AlistPath(PureAlistPath):
    """"""

    @cached_property
    def _client(self) -> Client:
        if self.drive == "":
            raise AlistError("当前对象没有设置server")

        try:
            _server = ALIST_SERVER_INFO[self.drive]
            return Client(
                _server.server,
                token=_server.token,
                **_server.kwargs,
            )
        except KeyError:
            raise AlistError(f"当前服务器[{self.drive}]尚未登陆")

    # def is_absolute(self) -> bool:
    def as_download_uri(self):
        if not self.is_absolute():
            raise ValueError("relative path can't be expressed as a file URI")
        if self.is_dir():
            raise IsADirectoryError()
        return self.drive + "/d" + self.as_posix() + "?sign=" + self.stat().sign

    @lru_cache()
    def stat(self) -> RawItem:
        _raw = self._client.get_item_info(self.as_posix())
        if _raw.code == 200:
            data = _raw.data
            return data
        if _raw.code == 500 and (
                "object not found" in _raw.message or "storage not found" in _raw.message
        ):
            raise FileNotFoundError(_raw.message)
        raise AlistError(_raw.message)

    def is_dir(self):
        """"""
        return self.stat().is_dir

    def is_file(self):
        """"""
        return not self.stat().is_dir

    def is_link(self):
        raise NotImplementedError("AlistPath不支持连接.")

    def exists(self):
        """"""
        try:
            return bool(self.stat())
        except FileNotFoundError:
            return False

    def iterdir(self) -> Iterator["AlistPath"]:
        """"""
        if not self.is_dir():
            raise

        for item in (
                self._client.list_files(self.as_posix(), refresh=True).data.content or []
        ):
            yield self.joinpath(item.name)

    def read_text(self):
        """"""
        return self._client.get(self.as_download_uri(), follow_redirects=True).text

    def read_bytes(self):
        """"""
        return self._client.get(self.as_download_uri(), follow_redirects=True).content

    def write_text(self, data: str, as_task=False):
        """"""
        return self.write_bytes(data.encode(), as_task=as_task)

    def write_bytes(self, data: bytes, as_task=False):
        """"""

        _res = self._client.upload_file_put(data, self.as_posix(), as_task=as_task)
        if _res.code == 200:
            return self.stat()
        return None



import json
import logging
from functools import wraps
from json import JSONDecodeError

import httpx
from pydantic import ValidationError

from alist_sdk.models import Resp, ListItem, Item, RawItem, BaseModel, TaskType

logger = logging.getLogger("alist-sdk.verify")

__all__ = [
    "Verify",
    "verify",
    "AsyncVerify",
    "async_verify",
]


class Verify:
    def __init__(self):
        self.locals: dict = {}
        self.request: httpx.Request | None = None

    def add_parent(self, res_dict: dict):
        res_dict.setdefault("parent", self.locals.get("path"))

    @property
    def ex_action(self):
        """"""
        return {
            ListItem: [self.add_parent],
            Item: [self.add_parent],
            RawItem: [self.add_parent],
            dict: [],
            BaseModel: [],
        }

    def acting(self, resp: Resp) -> Resp:
        """对Resp做额外的修饰"""
        if not isinstance(self.request, httpx.Request):
            return resp

        if (
            self.request.url.path
            in [
                *[
                    f"/api/admin/task/{tt}/{ts}"
                    for tt in TaskType
                    for ts in ["done", "undone"]
                ],
                "/user/info",
            ]
            and resp.code == 200
        ):
            resp.data = resp.data or []

        return resp

    def _verify(self, local_s, res: httpx.Response):
        self.locals.update(local_s)
        self.request = res.request
        url = res.request.url.path
        logger.debug(f"收到响应: {url}[{res.status_code}]  {res.text}")
        try:
            res_dict = res.json()
            resp = Resp.model_validate(res_dict)
            return self.acting(resp)

        except JSONDecodeError:
            logger.warning("JsonDecodeError: [http_status: %d] ", res.status_code)
            return Resp(
                code=res.status_code, message=f"JsonDecodeError: {res.text}", data=None
            )

        except ValidationError as _e:
            req_headers = "\n>>> ".join(
                f"{k}: {v}" for k, v in res.request.headers.items()
            )
            res_headers = "\n<<< ".join(f"{k}: {v}" for k, v in res.headers.items())
            logger.error(
                f"RESP验证错误：\n"
                f">>> {res.request.method} {res.request.url}"
                f">>> {req_headers}\n\n"
                f">>> {local_s = }\n"
                f"========= Request End =========\n"
                f"<<< {res_headers}\n\n"
                f"{json.dumps(res.json(), indent=2, ensure_ascii=False)} \n"
                f"========= Resp End =========",
            )
            raise _e

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return self._verify(*func(*args, **kwargs))

        return wrapper  # 返回函数


verify = Verify


class AsyncVerify(Verify):
    """"""

    def __call__(self, func):
        @wraps(func)
        async def async_wrapper(*args, **kwargs) -> Resp:
            return self._verify(*(await func(*args, **kwargs)))

        return async_wrapper  # 返回函数


async_verify = AsyncVerify



# -*- coding: utf-8 -*-

"""Object-oriented filesystem paths.

This module provides classes to represent abstract paths and concrete
paths with operations that have semantics appropriate for different
operating systems.
"""

import fnmatch
import functools
import io
import ntpath
import os
import posixpath
import re
import sys
import warnings
from _collections_abc import Sequence
from errno import ENOENT, ENOTDIR, EBADF, ELOOP
from stat import S_ISDIR, S_ISLNK, S_ISREG, S_ISSOCK, S_ISBLK, S_ISCHR, S_ISFIFO
from urllib.parse import quote_from_bytes as urlquote_from_bytes


__all__ = [
    "PurePath",
    "PurePosixPath",
    "PureWindowsPath",
    "Path",
    "PosixPath",
    "WindowsPath",
]

#
# Internals
#


def splitroot(p):
    """Split a pathname into drive, root and tail. On Posix, drive is always
    empty; the root may be empty, a single slash, or two slashes. The tail
    contains anything after the root. For example:

        splitroot('foo/bar') == ('', '', 'foo/bar')
        splitroot('/foo/bar') == ('', '/', 'foo/bar')
        splitroot('//foo/bar') == ('', '//', 'foo/bar')
        splitroot('///foo/bar') == ('', '/', '//foo/bar')
    """
    p = os.fspath(p)
    if isinstance(p, bytes):
        sep = b"/"
        empty = b""
    else:
        sep = "/"
        empty = ""
    if p[:1] != sep:
        # Relative path, e.g.: 'foo'
        return empty, empty, p
    elif p[1:2] != sep or p[2:3] == sep:
        # Absolute path, e.g.: '/foo', '///foo', '////foo', etc.
        return empty, sep, p[1:]
    else:
        # Precisely two leading slashes, e.g.: '//foo'. Implementation defined per POSIX, see
        # https://pubs.opengroup.org/onlinepubs/9699919799/basedefs/V1_chap04.html#tag_04_13
        return empty, p[:2], p[2:]


setattr(posixpath, "splitroot", splitroot)

# Reference for Windows paths can be found at
# https://learn.microsoft.com/en-gb/windows/win32/fileio/naming-a-file .
_WIN_RESERVED_NAMES = frozenset(
    {"CON", "PRN", "AUX", "NUL", "CONIN$", "CONOUT$"}
    | {f"COM{c}" for c in "123456789\xb9\xb2\xb3"}
    | {f"LPT{c}" for c in "123456789\xb9\xb2\xb3"}
)

_WINERROR_NOT_READY = 21  # drive exists but is not accessible
_WINERROR_INVALID_NAME = 123  # fix for bpo-35306
_WINERROR_CANT_RESOLVE_FILENAME = 1921  # broken symlink pointing to itself

# EBADF - guard against macOS `stat` throwing EBADF
_IGNORED_ERRNOS = (ENOENT, ENOTDIR, EBADF, ELOOP)

_IGNORED_WINERRORS = (
    _WINERROR_NOT_READY,
    _WINERROR_INVALID_NAME,
    _WINERROR_CANT_RESOLVE_FILENAME,
)


def _ignore_error(exception):
    return (
        getattr(exception, "errno", None) in _IGNORED_ERRNOS
        or getattr(exception, "winerror", None) in _IGNORED_WINERRORS
    )


@functools.cache
def _is_case_sensitive(flavour):
    return flavour.normcase("Aa") == "Aa"


#
# Globbing helpers
#


# fnmatch.translate() returns a regular expression that includes a prefix and
# a suffix, which enable matching newlines and ensure the end of the string is
# matched, respectively. These features are undesirable for our implementation
# of PurePatch.match(), which represents path separators as newlines and joins
# pattern segments together. As a workaround, we define a slice object that
# can remove the prefix and suffix from any translate() result. See the
# _compile_pattern_lines() function for more details.
_FNMATCH_PREFIX, _FNMATCH_SUFFIX = fnmatch.translate("_").split("_")
_FNMATCH_SLICE = slice(len(_FNMATCH_PREFIX), -len(_FNMATCH_SUFFIX))
_SWAP_SEP_AND_NEWLINE = {
    "/": str.maketrans({"/": "\n", "\n": "/"}),
    "\\": str.maketrans({"\\": "\n", "\n": "\\"}),
}


@functools.lru_cache()
def _make_selector(pattern_parts, flavour, case_sensitive):
    pat = pattern_parts[0]
    if not pat:
        return _TerminatingSelector()
    if pat == "**":
        child_parts_idx = 1
        while (
            child_parts_idx < len(pattern_parts)
            and pattern_parts[child_parts_idx] == "**"
        ):
            child_parts_idx += 1
        child_parts = pattern_parts[child_parts_idx:]
        if "**" in child_parts:
            cls = _DoubleRecursiveWildcardSelector
        else:
            cls = _RecursiveWildcardSelector
    else:
        child_parts = pattern_parts[1:]
        if pat == "..":
            cls = _ParentSelector
        elif "**" in pat:
            raise ValueError(
                "Invalid pattern: '**' can only be an entire path component"
            )
        else:
            cls = _WildcardSelector
    return cls(pat, child_parts, flavour, case_sensitive)


@functools.lru_cache(maxsize=256)
def _compile_pattern(pat, case_sensitive):
    flags = re.NOFLAG if case_sensitive else re.IGNORECASE
    return re.compile(fnmatch.translate(pat), flags).match


@functools.lru_cache()
def _compile_pattern_lines(pattern_lines, case_sensitive):
    """Compile the given pattern lines to an `re.Pattern` object.

    The *pattern_lines* argument is a glob-style pattern (e.g. '*/*.py') with
    its path separators and newlines swapped (e.g. '*\n*.py`). By using
    newlines to separate path components, and not setting `re.DOTALL`, we
    ensure that the `*` wildcard cannot match path separators.

    The returned `re.Pattern` object may have its `match()` method called to
    match a complete pattern, or `search()` to match from the right. The
    argument supplied to these methods must also have its path separators and
    newlines swapped.
    """

    # Match the start of the path, or just after a path separator
    parts = ["^"]
    for part in pattern_lines.splitlines(keepends=True):
        if part == "*\n":
            part = r".+\n"
        elif part == "*":
            part = r".+"
        else:
            # Any other component: pass to fnmatch.translate(). We slice off
            # the common prefix and suffix added by translate() to ensure that
            # re.DOTALL is not set, and the end of the string not matched,
            # respectively. With DOTALL not set, '*' wildcards will not match
            # path separators, because the '.' characters in the pattern will
            # not match newlines.
            part = fnmatch.translate(part)[_FNMATCH_SLICE]
        parts.append(part)
    # Match the end of the path, always.
    parts.append(r"\Z")
    flags = re.MULTILINE
    if not case_sensitive:
        flags |= re.IGNORECASE
    return re.compile("".join(parts), flags=flags)


class _Selector:
    """A selector matches a specific glob pattern part against the children
    of a given path."""

    def __init__(self, child_parts, flavour, case_sensitive):
        self.child_parts = child_parts
        if child_parts:
            self.successor = _make_selector(child_parts, flavour, case_sensitive)
            self.dironly = True
        else:
            self.successor = _TerminatingSelector()
            self.dironly = False

    def select_from(self, parent_path):
        """Iterate over all child paths of `parent_path` matched by this
        selector.  This can contain parent_path itself."""
        path_cls = type(parent_path)
        scandir = path_cls._scandir
        if not parent_path.is_dir():
            return iter([])
        return self._select_from(parent_path, scandir)


class _TerminatingSelector:

    def _select_from(self, parent_path, scandir):
        yield parent_path


class _ParentSelector(_Selector):

    def __init__(self, name, child_parts, flavour, case_sensitive):
        _Selector.__init__(self, child_parts, flavour, case_sensitive)

    def _select_from(self, parent_path, scandir):
        path = parent_path._make_child_relpath("..")
        for p in self.successor._select_from(path, scandir):
            yield p


class _WildcardSelector(_Selector):

    def __init__(self, pat, child_parts, flavour, case_sensitive):
        _Selector.__init__(self, child_parts, flavour, case_sensitive)
        if case_sensitive is None:
            # TODO: evaluate case-sensitivity of each directory in _select_from()
            case_sensitive = _is_case_sensitive(flavour)
        self.match = _compile_pattern(pat, case_sensitive)

    def _select_from(self, parent_path, scandir):
        try:
            # We must close the scandir() object before proceeding to
            # avoid exhausting file descriptors when globbing deep trees.
            with scandir(parent_path) as scandir_it:
                entries = list(scandir_it)
        except OSError:
            pass
        else:
            for entry in entries:
                if self.dironly:
                    try:
                        if not entry.is_dir():
                            continue
                    except OSError:
                        continue
                name = entry.name
                if self.match(name):
                    path = parent_path._make_child_relpath(name)
                    for p in self.successor._select_from(path, scandir):
                        yield p


class _RecursiveWildcardSelector(_Selector):

    def __init__(self, pat, child_parts, flavour, case_sensitive):
        _Selector.__init__(self, child_parts, flavour, case_sensitive)

    def _iterate_directories(self, parent_path):
        yield parent_path
        for dirpath, dirnames, _ in parent_path.walk():
            for dirname in dirnames:
                yield dirpath._make_child_relpath(dirname)

    def _select_from(self, parent_path, scandir):
        successor_select = self.successor._select_from
        for starting_point in self._iterate_directories(parent_path):
            for p in successor_select(starting_point, scandir):
                yield p


class _DoubleRecursiveWildcardSelector(_RecursiveWildcardSelector):
    """
    Like _RecursiveWildcardSelector, but also de-duplicates results from
    successive selectors. This is necessary if the pattern contains
    multiple non-adjacent '**' segments.
    """

    def _select_from(self, parent_path, scandir):
        yielded = set()
        try:
            for p in super()._select_from(parent_path, scandir):
                if p not in yielded:
                    yield p
                    yielded.add(p)
        finally:
            yielded.clear()


#
# Public API
#


class _PathParents(Sequence):
    """This object provides sequence-like access to the logical ancestors
    of a path.  Don't try to construct it yourself."""

    __slots__ = ("_path", "_drv", "_root", "_tail")

    def __init__(self, path):
        self._path = path
        self._drv = path.drive
        self._root = path.root
        self._tail = path._tail

    def __len__(self):
        return len(self._tail)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return tuple(self[i] for i in range(*idx.indices(len(self))))

        if idx >= len(self) or idx < -len(self):
            raise IndexError(idx)
        if idx < 0:
            idx += len(self)
        return self._path._from_parsed_parts(
            self._drv, self._root, self._tail[: -idx - 1]
        )

    def __repr__(self):
        return "<{}.parents>".format(type(self._path).__name__)


class PurePath(object):
    """Base class for manipulating paths without I/O.

    PurePath represents a filesystem path and offers operations which
    don't imply any actual filesystem I/O.  Depending on your system,
    instantiating a PurePath will return either a PurePosixPath or a
    PureWindowsPath object.  You can also instantiate either of these classes
    directly, regardless of your system.
    """

    __slots__ = (
        # The `_raw_paths` slot stores unnormalized string paths. This is set
        # in the `__init__()` method.
        "_raw_paths",
        # The `_drv`, `_root` and `_tail_cached` slots store parsed and
        # normalized parts of the path. They are set when any of the `drive`,
        # `root` or `_tail` properties are accessed for the first time. The
        # three-part division corresponds to the result of
        # `os.path.splitroot()`, except that the tail is further split on path
        # separators (i.e. it is a list of strings), and that the root and
        # tail are normalized.
        "_drv",
        "_root",
        "_tail_cached",
        # The `_str` slot stores the string representation of the path,
        # computed from the drive, root and tail when `__str__()` is called
        # for the first time. It's used to implement `_str_normcase`
        "_str",
        # The `_str_normcase_cached` slot stores the string path with
        # normalized case. It is set when the `_str_normcase` property is
        # accessed for the first time. It's used to implement `__eq__()`
        # `__hash__()`, and `_parts_normcase`
        "_str_normcase_cached",
        # The `_parts_normcase_cached` slot stores the case-normalized
        # string path after splitting on path separators. It's set when the
        # `_parts_normcase` property is accessed for the first time. It's used
        # to implement comparison methods like `__lt__()`.
        "_parts_normcase_cached",
        # The `_lines_cached` slot stores the string path with path separators
        # and newlines swapped. This is used to implement `match()`.
        "_lines_cached",
        # The `_hash` slot stores the hash of the case-normalized string
        # path. It's set when `__hash__()` is called for the first time.
        "_hash",
    )
    _flavour = os.path

    def __new__(cls, *args, **kwargs):
        """Construct a PurePath from one or several strings and or existing
        PurePath objects.  The strings and path objects are combined so as
        to yield a canonicalized path, which is incorporated into the
        new PurePath object.
        """
        if cls is PurePath:
            cls = PureWindowsPath if os.name == "nt" else PurePosixPath
        return object.__new__(cls)

    def __reduce__(self):
        # Using the parts tuple helps share interned path parts
        # when pickling related paths.
        return (self.__class__, self.parts)

    def __init__(self, *args):
        paths = []
        for arg in args:
            if isinstance(arg, PurePath):
                if arg._flavour is ntpath and self._flavour is posixpath:
                    # GH-103631: Convert separators for backwards compatibility.
                    paths.extend(path.replace("\\", "/") for path in arg._raw_paths)
                else:
                    paths.extend(arg._raw_paths)
            else:
                try:
                    path = os.fspath(arg)
                except TypeError:
                    path = arg
                if not isinstance(path, str):
                    raise TypeError(
                        "argument should be a str or an os.PathLike "
                        "object where __fspath__ returns a str, "
                        f"not {type(path).__name__!r}"
                    )
                paths.append(path)
        self._raw_paths = paths

    def with_segments(self, *pathsegments):
        """Construct a new path object from any number of path-like objects.
        Subclasses may override this method to customize how new path objects
        are created from methods like `iterdir()`.
        """
        return type(self)(*pathsegments)

    @classmethod
    def _parse_path(cls, path):
        if not path:
            return "", "", []
        sep = cls._flavour.sep
        altsep = cls._flavour.altsep
        if altsep:
            path = path.replace(altsep, sep)
        drv, root, rel = cls._flavour.splitroot(path)
        if not root and drv.startswith(sep) and not drv.endswith(sep):
            drv_parts = drv.split(sep)
            if len(drv_parts) == 4 and drv_parts[2] not in "?.":
                # e.g. //server/share
                root = sep
            elif len(drv_parts) == 6:
                # e.g. //?/unc/server/share
                root = sep
        parsed = [sys.intern(str(x)) for x in rel.split(sep) if x and x != "."]
        return drv, root, parsed

    def _load_parts(self):
        paths = self._raw_paths
        if len(paths) == 0:
            path = ""
        elif len(paths) == 1:
            path = paths[0]
        else:
            path = self._flavour.join(*paths)
        drv, root, tail = self._parse_path(path)
        self._drv = drv
        self._root = root
        self._tail_cached = tail

    def _from_parsed_parts(self, drv, root, tail):
        path_str = self._format_parsed_parts(drv, root, tail)
        path = self.with_segments(path_str)
        path._str = path_str or "."
        path._drv = drv
        path._root = root
        path._tail_cached = tail
        return path

    @classmethod
    def _format_parsed_parts(cls, drv, root, tail):
        if drv or root:
            return drv + root + cls._flavour.sep.join(tail)
        elif tail and cls._flavour.splitdrive(tail[0])[0]:
            tail = ["."] + tail
        return cls._flavour.sep.join(tail)

    def __str__(self):
        """Return the string representation of the path, suitable for
        passing to system calls."""
        try:
            return self._str
        except AttributeError:
            self._str = (
                self._format_parsed_parts(self.drive, self.root, self._tail) or "."
            )
            return self._str

    def __fspath__(self):
        return str(self)

    def as_posix(self):
        """Return the string representation of the path with forward (/)
        slashes."""
        f = self._flavour
        return str(self).replace(f.sep, "/")

    def __bytes__(self):
        """Return the bytes representation of the path.  This is only
        recommended to use under Unix."""
        return os.fsencode(self)

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.as_posix())

    def as_uri(self):
        """Return the path as a 'file' URI."""
        if not self.is_absolute():
            raise ValueError("relative path can't be expressed as a file URI")

        drive = self.drive
        if len(drive) == 2 and drive[1] == ":":
            # It's a path on a local drive => 'file:///c:/a/b'
            prefix = "file:///" + drive
            path = self.as_posix()[2:]
        elif drive:
            # It's a path on a network drive => 'file://host/share/a/b'
            prefix = "file:"
            path = self.as_posix()
        else:
            # It's a posix path => 'file:///etc/hosts'
            prefix = "file://"
            path = str(self)
        return prefix + urlquote_from_bytes(os.fsencode(path))

    @property
    def _str_normcase(self):
        # String with normalized case, for hashing and equality checks
        try:
            return self._str_normcase_cached
        except AttributeError:
            if _is_case_sensitive(self._flavour):
                self._str_normcase_cached = str(self)
            else:
                self._str_normcase_cached = str(self).lower()
            return self._str_normcase_cached

    @property
    def _parts_normcase(self):
        # Cached parts with normalized case, for comparisons.
        try:
            return self._parts_normcase_cached
        except AttributeError:
            self._parts_normcase_cached = self._str_normcase.split(self._flavour.sep)
            return self._parts_normcase_cached

    @property
    def _lines(self):
        # Path with separators and newlines swapped, for pattern matching.
        try:
            return self._lines_cached
        except AttributeError:
            path_str = str(self)
            if path_str == ".":
                self._lines_cached = ""
            else:
                trans = _SWAP_SEP_AND_NEWLINE[self._flavour.sep]
                self._lines_cached = path_str.translate(trans)
            return self._lines_cached

    def __eq__(self, other):
        if not isinstance(other, PurePath):
            return NotImplemented
        return (
            self._str_normcase == other._str_normcase
            and self._flavour is other._flavour
        )

    def __hash__(self):
        try:
            return self._hash
        except AttributeError:
            self._hash = hash(self._str_normcase)
            return self._hash

    def __lt__(self, other):
        if not isinstance(other, PurePath) or self._flavour is not other._flavour:
            return NotImplemented
        return self._parts_normcase < other._parts_normcase

    def __le__(self, other):
        if not isinstance(other, PurePath) or self._flavour is not other._flavour:
            return NotImplemented
        return self._parts_normcase <= other._parts_normcase

    def __gt__(self, other):
        if not isinstance(other, PurePath) or self._flavour is not other._flavour:
            return NotImplemented
        return self._parts_normcase > other._parts_normcase

    def __ge__(self, other):
        if not isinstance(other, PurePath) or self._flavour is not other._flavour:
            return NotImplemented
        return self._parts_normcase >= other._parts_normcase

    @property
    def drive(self):
        """The drive prefix (letter or UNC path), if any."""
        try:
            return self._drv
        except AttributeError:
            self._load_parts()
            return self._drv

    @property
    def root(self):
        """The root of the path, if any."""
        try:
            return self._root
        except AttributeError:
            self._load_parts()
            return self._root

    @property
    def _tail(self):
        try:
            return self._tail_cached
        except AttributeError:
            self._load_parts()
            return self._tail_cached

    @property
    def anchor(self):
        """The concatenation of the drive and root, or ''."""
        anchor = self.drive + self.root
        return anchor

    @property
    def name(self):
        """The final path component, if any."""
        tail = self._tail
        if not tail:
            return ""
        return tail[-1]

    @property
    def suffix(self):
        """
        The final component's last suffix, if any.

        This includes the leading period. For example: '.txt'
        """
        name = self.name
        i = name.rfind(".")
        if 0 < i < len(name) - 1:
            return name[i:]
        else:
            return ""

    @property
    def suffixes(self):
        """
        A list of the final component's suffixes, if any.

        These include the leading periods. For example: ['.tar', '.gz']
        """
        name = self.name
        if name.endswith("."):
            return []
        name = name.lstrip(".")
        return ["." + suffix for suffix in name.split(".")[1:]]

    @property
    def stem(self):
        """The final path component, minus its last suffix."""
        name = self.name
        i = name.rfind(".")
        if 0 < i < len(name) - 1:
            return name[:i]
        else:
            return name

    def with_name(self, name):
        """Return a new path with the file name changed."""
        if not self.name:
            raise ValueError("%r has an empty name" % (self,))
        f = self._flavour
        if not name or f.sep in name or (f.altsep and f.altsep in name) or name == ".":
            raise ValueError("Invalid name %r" % (name))
        return self._from_parsed_parts(self.drive, self.root, self._tail[:-1] + [name])

    def with_stem(self, stem):
        """Return a new path with the stem changed."""
        return self.with_name(stem + self.suffix)

    def with_suffix(self, suffix):
        """Return a new path with the file suffix changed.  If the path
        has no suffix, add given suffix.  If the given suffix is an empty
        string, remove the suffix from the path.
        """
        f = self._flavour
        if f.sep in suffix or f.altsep and f.altsep in suffix:
            raise ValueError("Invalid suffix %r" % (suffix,))
        if suffix and not suffix.startswith(".") or suffix == ".":
            raise ValueError("Invalid suffix %r" % (suffix))
        name = self.name
        if not name:
            raise ValueError("%r has an empty name" % (self,))
        old_suffix = self.suffix
        if not old_suffix:
            name = name + suffix
        else:
            name = name[: -len(old_suffix)] + suffix
        return self._from_parsed_parts(self.drive, self.root, self._tail[:-1] + [name])

    def relative_to(self, other, /, *_deprecated, walk_up=False):
        """Return the relative path to another path identified by the passed
        arguments.  If the operation is not possible (because this is not
        related to the other path), raise ValueError.

        The *walk_up* parameter controls whether `..` may be used to resolve
        the path.
        """
        if _deprecated:
            msg = (
                "support for supplying more than one positional argument "
                "to pathlib.PurePath.relative_to() is deprecated and "
                "scheduled for removal in Python {remove}"
            )
            warnings._deprecated(
                "pathlib.PurePath.relative_to(*args)", msg, remove=(3, 14)
            )
        other = self.with_segments(other, *_deprecated)
        for step, path in enumerate([other] + list(other.parents)):
            if self.is_relative_to(path):
                break
            elif not walk_up:
                raise ValueError(
                    f"{str(self)!r} is not in the subpath of {str(other)!r}"
                )
            elif path.name == "..":
                raise ValueError(f"'..' segment in {str(other)!r} cannot be walked")
        else:
            raise ValueError(f"{str(self)!r} and {str(other)!r} have different anchors")
        parts = [".."] * step + self._tail[len(path._tail) :]
        return self.with_segments(*parts)

    def is_relative_to(self, other, /, *_deprecated):
        """Return True if the path is relative to another path or False."""
        if _deprecated:
            msg = (
                "support for supplying more than one argument to "
                "pathlib.PurePath.is_relative_to() is deprecated and "
                "scheduled for removal in Python {remove}"
            )
            warnings._deprecated(
                "pathlib.PurePath.is_relative_to(*args)", msg, remove=(3, 14)
            )
        other = self.with_segments(other, *_deprecated)
        return other == self or other in self.parents

    @property
    def parts(self):
        """An object providing sequence-like access to the
        components in the filesystem path."""
        if self.drive or self.root:
            return (self.drive + self.root,) + tuple(self._tail)
        else:
            return tuple(self._tail)

    def joinpath(self, *pathsegments):
        """Combine this path with one or several arguments, and return a
        new path representing either a subpath (if all arguments are relative
        paths) or a totally different path (if one of the arguments is
        anchored).
        """
        return self.with_segments(self, *pathsegments)

    def __truediv__(self, key):
        try:
            return self.joinpath(key)
        except TypeError:
            return NotImplemented

    def __rtruediv__(self, key):
        try:
            return self.with_segments(key, self)
        except TypeError:
            return NotImplemented

    @property
    def parent(self):
        """The logical parent of the path."""
        drv = self.drive
        root = self.root
        tail = self._tail
        if not tail:
            return self
        return self._from_parsed_parts(drv, root, tail[:-1])

    @property
    def parents(self):
        """A sequence of this path's logical parents."""
        # The value of this property should not be cached on the path object,
        # as doing so would introduce a reference cycle.
        return _PathParents(self)

    def is_absolute(self):
        """True if the path is absolute (has both a root and, if applicable,
        a drive)."""
        if self._flavour is ntpath:
            # ntpath.isabs() is defective - see GH-44626.
            return bool(self.drive and self.root)
        elif self._flavour is posixpath:
            # Optimization: work with raw paths on POSIX.
            for path in self._raw_paths:
                if path.startswith("/"):
                    return True
            return False
        else:
            return self._flavour.isabs(str(self))

    def is_reserved(self):
        """Return True if the path contains one of the special names reserved
        by the system, if any."""
        if self._flavour is posixpath or not self._tail:
            return False

        # NOTE: the rules for reserved names seem somewhat complicated
        # (e.g. r"..\NUL" is reserved but not r"foo\NUL" if "foo" does not
        # exist). We err on the side of caution and return True for paths
        # which are not considered reserved by Windows.
        if self.drive.startswith("\\\\"):
            # UNC paths are never reserved.
            return False
        name = self._tail[-1].partition(".")[0].partition(":")[0].rstrip(" ")
        return name.upper() in _WIN_RESERVED_NAMES

    def match(self, path_pattern, *, case_sensitive=None):
        """
        Return True if this path matches the given pattern.
        """
        if not isinstance(path_pattern, PurePath):
            path_pattern = self.with_segments(path_pattern)
        if case_sensitive is None:
            case_sensitive = _is_case_sensitive(self._flavour)
        pattern = _compile_pattern_lines(path_pattern._lines, case_sensitive)
        if path_pattern.drive or path_pattern.root:
            return pattern.match(self._lines) is not None
        elif path_pattern._tail:
            return pattern.search(self._lines) is not None
        else:
            raise ValueError("empty pattern")


# Can't subclass os.PathLike from PurePath and keep the constructor
# optimizations in PurePath.__slots__.
os.PathLike.register(PurePath)


class PurePosixPath(PurePath):
    """PurePath subclass for non-Windows systems.

    On a POSIX system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """

    _flavour = posixpath
    __slots__ = ()


class PureWindowsPath(PurePath):
    """PurePath subclass for Windows systems.

    On a Windows system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """

    _flavour = ntpath
    __slots__ = ()


# Filesystem-accessing classes


class Path(PurePath):
    """PurePath subclass that can make system calls.

    Path represents a filesystem path but unlike PurePath, also offers
    methods to do system calls on path objects. Depending on your system,
    instantiating a Path will return either a PosixPath or a WindowsPath
    object. You can also instantiate a PosixPath or WindowsPath directly,
    but cannot instantiate a WindowsPath on a POSIX system or vice versa.
    """

    __slots__ = ()

    def stat(self, *, follow_symlinks=True):
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        return os.stat(self, follow_symlinks=follow_symlinks)

    def lstat(self):
        """
        Like stat(), except if the path points to a symlink, the symlink's
        status information is returned, rather than its target's.
        """
        return self.stat(follow_symlinks=False)

    # Convenience functions for querying the stat results

    def exists(self, *, follow_symlinks=True):
        """
        Whether this path exists.

        This method normally follows symlinks; to check whether a symlink exists,
        add the argument follow_symlinks=False.
        """
        try:
            self.stat(follow_symlinks=follow_symlinks)
        except OSError as e:
            if not _ignore_error(e):
                raise
            return False
        except ValueError:
            # Non-encodable path
            return False
        return True

    def is_dir(self):
        """
        Whether this path is a directory.
        """
        try:
            return S_ISDIR(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_file(self):
        """
        Whether this path is a regular file (also True for symlinks pointing
        to regular files).
        """
        try:
            return S_ISREG(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_mount(self):
        """
        Check if this path is a mount point
        """
        return self._flavour.ismount(self)

    def is_symlink(self):
        """
        Whether this path is a symbolic link.
        """
        try:
            return S_ISLNK(self.lstat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_junction(self):
        """
        Whether this path is a junction.
        """
        return self._flavour.isjunction(self)

    def is_block_device(self):
        """
        Whether this path is a block device.
        """
        try:
            return S_ISBLK(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_char_device(self):
        """
        Whether this path is a character device.
        """
        try:
            return S_ISCHR(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_fifo(self):
        """
        Whether this path is a FIFO.
        """
        try:
            return S_ISFIFO(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_socket(self):
        """
        Whether this path is a socket.
        """
        try:
            return S_ISSOCK(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def samefile(self, other_path):
        """Return whether other_path is the same or not as this file
        (as returned by os.path.samefile()).
        """
        st = self.stat()
        try:
            other_st = other_path.stat()
        except AttributeError:
            other_st = self.with_segments(other_path).stat()
        return self._flavour.samestat(st, other_st)

    def open(self, mode="r", buffering=-1, encoding=None, errors=None, newline=None):
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        if "b" not in mode:
            encoding = io.text_encoding(encoding)
        return io.open(self, mode, buffering, encoding, errors, newline)

    def read_bytes(self):
        """
        Open the file in bytes mode, read it, and close the file.
        """
        with self.open(mode="rb") as f:
            return f.read()

    def read_text(self, encoding=None, errors=None):
        """
        Open the file in text mode, read it, and close the file.
        """
        encoding = io.text_encoding(encoding)
        with self.open(mode="r", encoding=encoding, errors=errors) as f:
            return f.read()

    def write_bytes(self, data):
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        # type-check for the buffer interface before truncating the file
        view = memoryview(data)
        with self.open(mode="wb") as f:
            return f.write(view)

    def write_text(self, data, encoding=None, errors=None, newline=None):
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str):
            raise TypeError("data must be str, not %s" % data.__class__.__name__)
        encoding = io.text_encoding(encoding)
        with self.open(
            mode="w", encoding=encoding, errors=errors, newline=newline
        ) as f:
            return f.write(data)

    def iterdir(self):
        """Yield path objects of the directory contents.

        The children are yielded in arbitrary order, and the
        special entries '.' and '..' are not included.
        """
        for name in os.listdir(self):
            yield self._make_child_relpath(name)

    def _scandir(self):
        # bpo-24132: a future version of pathlib will support subclassing of
        # pathlib.Path to customize how the filesystem is accessed. This
        # includes scandir(), which is used to implement glob().
        return os.scandir(self)

    def _make_child_relpath(self, name):
        path_str = str(self)
        tail = self._tail
        if tail:
            path_str = f"{path_str}{self._flavour.sep}{name}"
        elif path_str != ".":
            path_str = f"{path_str}{name}"
        else:
            path_str = name
        path = self.with_segments(path_str)
        path._str = path_str
        path._drv = self.drive
        path._root = self.root
        path._tail_cached = tail + [name]
        return path

    def glob(self, pattern, *, case_sensitive=None):
        """Iterate over this subtree and yield all existing files (of any
        kind, including directories) matching the given relative pattern.
        """
        sys.audit("pathlib.Path.glob", self, pattern)
        if not pattern:
            raise ValueError("Unacceptable pattern: {!r}".format(pattern))
        drv, root, pattern_parts = self._parse_path(pattern)
        if drv or root:
            raise NotImplementedError("Non-relative patterns are unsupported")
        if pattern[-1] in (self._flavour.sep, self._flavour.altsep):
            pattern_parts.append("")
        selector = _make_selector(tuple(pattern_parts), self._flavour, case_sensitive)
        for p in selector.select_from(self):
            yield p

    def rglob(self, pattern, *, case_sensitive=None):
        """Recursively yield all existing files (of any kind, including
        directories) matching the given relative pattern, anywhere in
        this subtree.
        """
        sys.audit("pathlib.Path.rglob", self, pattern)
        drv, root, pattern_parts = self._parse_path(pattern)
        if drv or root:
            raise NotImplementedError("Non-relative patterns are unsupported")
        if pattern and pattern[-1] in (self._flavour.sep, self._flavour.altsep):
            pattern_parts.append("")
        selector = _make_selector(
            ("**",) + tuple(pattern_parts), self._flavour, case_sensitive
        )
        for p in selector.select_from(self):
            yield p

    def walk(self, top_down=True, on_error=None, follow_symlinks=False):
        """Walk the directory tree from this directory, similar to os.walk()."""
        sys.audit("pathlib.Path.walk", self, on_error, follow_symlinks)
        paths = [self]

        while paths:
            path = paths.pop()
            if isinstance(path, tuple):
                yield path
                continue

            # We may not have read permission for self, in which case we can't
            # get a list of the files the directory contains. os.walk()
            # always suppressed the exception in that instance, rather than
            # blow up for a minor reason when (say) a thousand readable
            # directories are still left to visit. That logic is copied here.
            try:
                scandir_it = path._scandir()
            except OSError as error:
                if on_error is not None:
                    on_error(error)
                continue

            with scandir_it:
                dirnames = []
                filenames = []
                for entry in scandir_it:
                    try:
                        is_dir = entry.is_dir(follow_symlinks=follow_symlinks)
                    except OSError:
                        # Carried over from os.path.isdir().
                        is_dir = False

                    if is_dir:
                        dirnames.append(entry.name)
                    else:
                        filenames.append(entry.name)

            if top_down:
                yield path, dirnames, filenames
            else:
                paths.append((path, dirnames, filenames))

            paths += [path._make_child_relpath(d) for d in reversed(dirnames)]

    def __init__(self, *args, **kwargs):
        if kwargs:
            msg = (
                "support for supplying keyword arguments to pathlib.PurePath "
                "is deprecated and scheduled for removal in Python {remove}"
            )
            warnings._deprecated("pathlib.PurePath(**kwargs)", msg, remove=(3, 14))
        super().__init__(*args)

    def __new__(cls, *args, **kwargs):
        if cls is Path:
            cls = WindowsPath if os.name == "nt" else PosixPath
        return object.__new__(cls)

    def __enter__(self):
        # In previous versions of pathlib, __exit__() marked this path as
        # closed; subsequent attempts to perform I/O would raise an IOError.
        # This functionality was never documented, and had the effect of
        # making Path objects mutable, contrary to PEP 428.
        # In Python 3.9 __exit__() was made a no-op.
        # In Python 3.11 __enter__() began emitting DeprecationWarning.
        # In Python 3.13 __enter__() and __exit__() should be removed.
        warnings.warn(
            "pathlib.Path.__enter__() is deprecated and scheduled "
            "for removal in Python 3.13; Path objects as a context "
            "manager is a no-op",
            DeprecationWarning,
            stacklevel=2,
        )
        return self

    def __exit__(self, t, v, tb):
        pass

    # Public API

    @classmethod
    def cwd(cls):
        """Return a new path pointing to the current working directory."""
        # We call 'absolute()' rather than using 'os.getcwd()' directly to
        # enable users to replace the implementation of 'absolute()' in a
        # subclass and benefit from the new behaviour here. This works because
        # os.path.abspath('.') == os.getcwd().
        return cls().absolute()

    @classmethod
    def home(cls):
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        return cls("~").expanduser()

    def absolute(self):
        """Return an absolute version of this path by prepending the current
        working directory. No normalization or symlink resolution is performed.

        Use resolve() to get the canonical path to a file.
        """
        if self.is_absolute():
            return self
        elif self.drive:
            # There is a CWD on each drive-letter drive.
            cwd = self._flavour.abspath(self.drive)
        else:
            cwd = os.getcwd()
            # Fast path for "empty" paths, e.g. Path("."), Path("") or Path().
            # We pass only one argument to with_segments() to avoid the cost
            # of joining, and we exploit the fact that getcwd() returns a
            # fully-normalized string by storing it in _str. This is used to
            # implement Path.cwd().
            if not self.root and not self._tail:
                result = self.with_segments(cwd)
                result._str = cwd
                return result
        return self.with_segments(cwd, self)

    def resolve(self, strict=False):
        """
        Make the path absolute, resolving all symlinks on the way and also
        normalizing it.
        """

        def check_eloop(e):
            winerror = getattr(e, "winerror", 0)
            if e.errno == ELOOP or winerror == _WINERROR_CANT_RESOLVE_FILENAME:
                raise RuntimeError("Symlink loop from %r" % e.filename)

        try:
            s = self._flavour.realpath(self, strict=strict)
        except OSError as e:
            check_eloop(e)
            raise
        p = self.with_segments(s)

        # In non-strict mode, realpath() doesn't raise on symlink loops.
        # Ensure we get an exception by calling stat()
        if not strict:
            try:
                p.stat()
            except OSError as e:
                check_eloop(e)
        return p

    def owner(self):
        """
        Return the login name of the file owner.
        """
        try:
            import pwd

            return pwd.getpwuid(self.stat().st_uid).pw_name
        except ImportError:
            raise NotImplementedError("Path.owner() is unsupported on this system")

    def group(self):
        """
        Return the group name of the file gid.
        """

        try:
            import grp

            return grp.getgrgid(self.stat().st_gid).gr_name
        except ImportError:
            raise NotImplementedError("Path.group() is unsupported on this system")

    def readlink(self):
        """
        Return the path to which the symbolic link points.
        """
        if not hasattr(os, "readlink"):
            raise NotImplementedError("os.readlink() not available on this system")
        return self.with_segments(os.readlink(self))

    def touch(self, mode=0o666, exist_ok=True):
        """
        Create this file with the given access mode, if it doesn't exist.
        """

        if exist_ok:
            # First try to bump modification time
            # Implementation note: GNU touch uses the UTIME_NOW option of
            # the utimensat() / futimens() functions.
            try:
                os.utime(self, None)
            except OSError:
                # Avoid exception chaining
                pass
            else:
                return
        flags = os.O_CREAT | os.O_WRONLY
        if not exist_ok:
            flags |= os.O_EXCL
        fd = os.open(self, flags, mode)
        os.close(fd)

    def mkdir(self, mode=0o777, parents=False, exist_ok=False):
        """
        Create a new directory at this given path.
        """
        try:
            os.mkdir(self, mode)
        except FileNotFoundError:
            if not parents or self.parent == self:
                raise
            self.parent.mkdir(parents=True, exist_ok=True)
            self.mkdir(mode, parents=False, exist_ok=exist_ok)
        except OSError:
            # Cannot rely on checking for EEXIST, since the operating system
            # could give priority to other errors like EACCES or EROFS
            if not exist_ok or not self.is_dir():
                raise

    def chmod(self, mode, *, follow_symlinks=True):
        """
        Change the permissions of the path, like os.chmod().
        """
        os.chmod(self, mode, follow_symlinks=follow_symlinks)

    def lchmod(self, mode):
        """
        Like chmod(), except if the path points to a symlink, the symlink's
        permissions are changed, rather than its target's.
        """
        self.chmod(mode, follow_symlinks=False)

    def unlink(self, missing_ok=False):
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        try:
            os.unlink(self)
        except FileNotFoundError:
            if not missing_ok:
                raise

    def rmdir(self):
        """
        Remove this directory.  The directory must be empty.
        """
        os.rmdir(self)

    def rename(self, target):
        """
        Rename this path to the target path.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        os.rename(self, target)
        return self.with_segments(target)

    def replace(self, target):
        """
        Rename this path to the target path, overwriting if that path exists.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        os.replace(self, target)
        return self.with_segments(target)

    def symlink_to(self, target, target_is_directory=False):
        """
        Make this path a symlink pointing to the target path.
        Note the order of arguments (link, target) is the reverse of os.symlink.
        """
        if not hasattr(os, "symlink"):
            raise NotImplementedError("os.symlink() not available on this system")
        os.symlink(target, self, target_is_directory)

    def hardlink_to(self, target):
        """
        Make this path a hard link pointing to the same file as *target*.

        Note the order of arguments (self, target) is the reverse of os.link's.
        """
        if not hasattr(os, "link"):
            raise NotImplementedError("os.link() not available on this system")
        os.link(target, self)

    def expanduser(self):
        """Return a new path with expanded ~ and ~user constructs
        (as returned by os.path.expanduser)
        """
        if not (self.drive or self.root) and self._tail and self._tail[0][:1] == "~":
            homedir = self._flavour.expanduser(self._tail[0])
            if homedir[:1] == "~":
                raise RuntimeError("Could not determine home directory.")
            drv, root, tail = self._parse_path(homedir)
            return self._from_parsed_parts(drv, root, tail + self._tail[1:])

        return self


class PosixPath(Path, PurePosixPath):
    """Path subclass for non-Windows systems.

    On a POSIX system, instantiating a Path should return this object.
    """

    __slots__ = ()

    if os.name == "nt":

        def __new__(cls, *args, **kwargs):
            raise NotImplementedError(
                f"cannot instantiate {cls.__name__!r} on your system"
            )


class WindowsPath(Path, PureWindowsPath):
    """Path subclass for Windows systems.

    On a Windows system, instantiating a Path should return this object.
    """

    __slots__ = ()

    if os.name != "nt":

        def __new__(cls, *args, **kwargs):
            raise NotImplementedError(
                f"cannot instantiate {cls.__name__!r} on your system"
            )



"""

0.30.4:
    path_lib.AlistPath.write_* 支持as_task参数

0.30.5:
    path_lib.AlistPath.read_* 自动跟随重定向，确保读取真实文件内容
    path_lib.AlistPath.as_download_uri() 返回添加签名

0.30.6:
    path_lib.py 重构，以支持3.12
    将AlistPath* 添加到 __init__.py 中。

0.30.7:
    AlistPath 实现了 mkdir, touch, unlink 方法, 并测试
    BUGFIX: client remove 方法参数错误
    path_lib 添加AlistPathType 以适配 Pydantic 。
    验证器现在将 /api/admin/task/ 返回None的data转换为[]

0.30.8 & 0.30.9:
    AlistPath: 实现relative_to -> str
    AlistPath: 实现rename
    pydantic类型支持： AbsAlistPathType, AlistPathType
    AlistPath: 实现re_stat() 以强制刷新缓存
    测试的alist端口更新为5245，防止多个服务冲突。
    AlistPath 添加构造器 - from_client

0.30.10:
    1. 现在AlistPath.write_bytes() 支持接收Path对象，读取文件并写入远程
    2. UPDATE:path_lib.login_server() 防止重复登陆
    3. BUGFIX: AlistPath.exists() 现在引用 re_stat()，以检查最新的状态
    4. AlistPath 允许递归创建目录
    5. AlistPath.as_download_uri -> get_download_uri, 获取方式更新。
    6. BUGFIX: 获取下载URL错误
    7. 客户端添加方法：admin_storage_update, admin_storage_delete
    8. 现在AlistPath.stat(force=False, retry=1, timeout=0.1): 在默认情况下获取不到文件,对象将会在0.1秒后重试一次.
    9. 现在AlistPath.re_stat(retry=1, timeout=1) == stat(True, 1, 1)
    10. 现在 tools.config.import_config 将忽略本地存储和任何key=id
    11. 当新文件出现在新的存储器中的时候,需要强制刷新list


0.30.11:
    1. BUGFIX: 从其他站点导入配置时，数据模型错误。
    2. BUGFIX: http://localhost:5244 无法正常识别
    3. UPDATE: 更新AplistPath.__repl__方法.
    4. UPDATE: 更新AlistPath.添加新的方法 set_stat，可以自定义设置stat属性，加快速度。
    5. UPDATE: 更新AlistPath.iterdir，在迭代时添加stat数据，加快速度。
    6. UPDATE: 更新AlistPath.stat，不再使用/api/fs/get 接口
    7. UPDATE: 限制alist_sdk.Client 和 alist_sdk.AsyncClient在多线程和协程中的并发量，默认30.


0.32.12:
    1. 添加Client.dict_files_items方法，以获取目录下的文件列表。
    2. 解决AListPath使用URL编码问题，现在可以正常使用中文路径。
    3. 修复AlistPath.stat() & AlistPath.raw_stat() 的缓存问题。
    4. 修复tools.config.import_configs_from_dict无法忽略本地存储的错误。
    5. Client Put AsTask 修复
    6. Verify 日志优化

0.36.13:
    1. alist版本支持到3.36.0
    2. 现在可以使用AlistPath(path, username="", password="", token="")的方式快速登录。
    3. 登录失败现在抛出异常。
    4. #3 Bugfix 为models中的全部可选字段添加默认值。
"""

__version__ = "0.36.13a4"

ALIST_VERSION = "v3.36.0"



import datetime
import logging
import os
from pathlib import PurePosixPath
from typing import Optional, Literal, Any

from pydantic import (
    BaseModel as _BaseModel,
    computed_field,
    field_serializer,
)

logger = logging.getLogger("alist-sdk.fs.model")

__all__ = [
    "SearchScopeModify",
    "TaskType",
    "TaskTypeModify",
    "TaskStateModify",
    "TaskStatusModify",
    "OrderDirectionModify",
    "OrderByModify",
    "ExtractFolderModify",
    "BaseModel",
    "Item",
    "RawItem",
    "ListItem",
    "DirItem",
    "SearchItem",
    "Me",
    "Task",
    "ListTask",
    "AsTask",
    "Resp",
    "HashInfo",
    "NoneType",
    "ListContents",
    "User",
    "Storage",
    "Meta",
    "Setting",
]

NoneType = type(None)

SearchScopeModify = Literal[0, 1, 2]  # 0-全部 1-文件夹 2-文件

TaskType = [
    "copy",
    "upload",
    "offline_download",
    "offline_download_transfer",
]
TaskTypeModify = Literal[
    "copy",
    "upload",
    "offline_download",
    "offline_download_transfer",
]
TaskStateModify = Literal[0, 1, 2, 3, 4, 5, 6, 7]
TaskStatusModify = Literal[
    "",
    "waiting",
    "running",
    "success",
    "failed",
    "getting src object",
    "uploading",
]

OrderDirectionModify = Literal["", "asc", "desc"]
OrderByModify = Literal["", "size", "name"]
ExtractFolderModify = Literal["", "front", "back", "none"]


class BaseModel(_BaseModel):
    parent: Optional[str | PurePosixPath | None] = ""

    @computed_field()
    @property
    def full_name(self) -> PurePosixPath:
        return PurePosixPath(self.parent).joinpath(self.name)

    @field_serializer("full_name", mode="wrap")
    def serializer_path(self, value: PurePosixPath, _) -> str:
        return value.as_posix()


class ListItem(_BaseModel):
    """列出的目录"""

    content: list["Item"] | None
    total: int  # 总数
    readme: str  # 用于渲染Readme
    header: Optional[str] = ""  # 用于渲染Header
    write: bool  # 是否可写
    provider: str


class HashInfo(_BaseModel):
    """Hash 信息"""

    md5: Optional[str] = None
    sha1: Optional[str] = None


class Item(BaseModel):
    """文件对象"""

    name: str  # 文件名
    size: int  # 文件大小
    is_dir: bool  # 是否是目录
    hashinfo: Optional[str] = "null"  # v3.29.0
    hash_info: Optional[HashInfo | None] = None  # v3.29.0
    modified: datetime.datetime  # 修改时间
    created: Optional[datetime.datetime] = None  # v3.29.0 创建时间
    sign: str  # 签名
    thumb: str  # 缩略图
    type: int  # 类型


class RawItem(Item):
    """一个对象的全部信息"""

    raw_url: str = ""
    readme: str = ""
    provider: str = ""
    related: str | None | list["RawItem"] = None


class SearchItem(BaseModel):
    parent: str
    name: str
    is_dir: bool
    size: int
    type: int


class DirItem(BaseModel):
    name: str
    modified: datetime.datetime


class Me(_BaseModel):
    """
    /api/me
    """

    id: int
    username: str
    password: str | None
    base_path: str
    role: int
    disabled: bool
    permission: int
    sso_id: str | None
    otp: Optional[bool] = False


class Task(_BaseModel):
    """
    »» id	string	false	none	id	none
    »» name	string	false	none	任务名	none
    »» state	string	false	none	任务完成状态	none
    »» status	string	false	none		none
    »» progress	integer	false	none	进度	none
    »» error	string	false	none	错误信息	none
    {
    "id":"0Vdzdi3BO_8vYujh72OTu",
    "name":"copy [/local](/test.txt) to [/local_dst](/)",
    "state":0,
    "status":"",
    "progress":0,
    "error":""
    }
    """

    id: str  # 任务ID
    name: str  # 任务名
    state: TaskStateModify  # 任务完成状态
    status: TaskStatusModify  # 任务状态
    progress: int | float  # 进度
    error: str  # 错误信息


class AsTask(_BaseModel):
    """/api/fs/put .headers"""

    task: Task


class ListTask(_BaseModel):
    tasks: list[Task]


class ID(_BaseModel):
    id: int | str | None


class Setting(_BaseModel):
    """/api/admin/setting/list .data.[]"""

    key: str
    value: Any
    help: str  # 帮助信息
    type: str  # TODO 类型指导
    options: str
    group: int
    flag: int


class Storage(_BaseModel):
    # /api/admin/storage/list .data.content.[]
    # ListContent.content[Storage]
    id: Optional[int] = None
    mount_path: str | os.PathLike
    order: int
    driver: str
    cache_expiration: int
    status: str
    addition: str  # TODO JSON
    remark: str = ""
    modified: datetime.datetime
    disabled: bool
    enable_sign: bool
    order_by: OrderByModify
    order_direction: OrderDirectionModify
    extract_folder: ExtractFolderModify  # 提取文件夹位置
    web_proxy: bool
    webdav_policy: str  # webdav策略
    down_proxy_url: str = ""


class User(_BaseModel):
    """/api/admin/user/list .data.content.[]"""

    id: int
    username: str
    password: str
    base_path: PurePosixPath = PurePosixPath("/")
    role: int  #
    disabled: bool
    permission: int
    sso_id: str


class Meta(_BaseModel):
    """/api/admin/meta/list .data.content.[]"""

    id: int
    path: PurePosixPath
    password: str
    p_sub: bool
    write: bool
    w_sub: bool
    hide: str
    h_sub: bool
    readme: str
    r_sub: bool
    header: str
    header_sub: bool


class ListContents(_BaseModel):
    content: list[SearchItem | Storage | User | Meta] | None
    total: int


class Resp(_BaseModel):
    code: int
    message: str
    # data: ListTask
    data: (
        None
        | str
        | Me
        | list[DirItem]
        | list[Setting]
        | ListItem
        | ListContents
        | RawItem
        | ListTask
        | list[Task]
        | AsTask
        | ID
    )



import json
import logging
from pathlib import Path

from alist_sdk.models import Resp, Storage
from alist_sdk.client import Client
from alist_sdk.tools.models import Configs

logger = logging.getLogger("alist-sdk.tools")

__all__ = [
    "import_configs_from_dict",
    "import_configs_from_json_str",
    "import_configs_from_json_file",
    "export_configs",
    "export_configs_to_dict",
    "export_configs_to_json",
]


def __printf(key, data, res: Resp):
    """记录"""
    name = (
        ""
        if isinstance(data, list)
        else (data.get("mount_path") or data.get("path") or data.get("username"))
    )
    if res.code == 200:
        print(f"created {key} [{name}]")
    else:
        print(f"Error: {key} [{name}]: {res.code}: {res.message}")


def import_configs_from_dict(client: Client, configs: dict):
    """从JSON中导入配置"""
    apis = {
        "settings": "/api/admin/setting/save",
        "users": "/api/admin/user/create",
        "storages": "/api/admin/storage/create",
        "metas": "/api/admin/meta/create",
    }

    for k, vs in configs.items():
        if k == "settings":
            res = client.verify_request("POST", apis[k], json=vs)
            __printf(k, vs, res)
            continue
        else:
            for v in vs:
                v: dict
                if k == "storages" and v.get("driver", "") == "Local":
                    continue
                v.pop("id", None)
                res = client.verify_request("POST", apis[k], json=v)
                __printf(k, v, res)


def import_configs_from_json_str(client, json_data: str):
    """"""
    return import_configs_from_dict(client, json.loads(json_data))


def import_configs_from_json_file(client, json_file):
    """"""
    if not Path(json_file).exists():
        raise FileNotFoundError()
    return import_configs_from_json_str(client, Path(json_file).read_text())


def export_configs(client: Client) -> Configs:
    """"""
    apis = {
        "settings": client.admin_setting_list().data,
        "users": client.admin_user_list().data.content,
        "storages": client.admin_storage_list().data.content,
        "metas": client.admin_meta_list().data.content,
    }
    return Configs(**apis)


def export_configs_to_dict(client: Client, *args, **kwargs) -> dict:
    return export_configs(client).model_dump(*args, **kwargs)


def export_configs_to_json(client: Client, *args, **kwargs) -> str:
    return export_configs(client).model_dump_json(*args, **kwargs)


def copy_configs(source_client: Client, target_client: Client):
    """"""
    return import_configs_from_dict(
        target_client,
        export_configs_to_dict(source_client),
    )


def copy_storages(source_client: Client, target_client: Client, *storage_names):
    """"""
    storages: list[Storage] = source_client.admin_storage_list().data.content
    if not storage_names:
        storage_names = [_.mount_path.strip("/") for _ in storages]
    for _s in storages:
        if _s.mount_path.strip("/") not in storage_names:
            continue
        target_client.admin_storage_create(_s)



from datetime import datetime
from pathlib import PurePosixPath, Path
import json

import pytest

from alist_sdk.models import *
from alist_sdk import models

MODEL_SIMPLE = Path(__file__).parent.joinpath("models_simple")


class TestModel:
    @pytest.mark.parametrize(
        "item",
        json.loads(MODEL_SIMPLE.joinpath("Items.json").read_text()),
        ids=lambda item: item.get("name"),
    )
    def test_item(self, item):
        a = Item(**item)
        assert isinstance(a.full_name, PurePosixPath)
        assert a.full_name.as_posix()
        assert isinstance(a.model_dump_json(), str)
        assert isinstance(a.created, datetime)
        print(a)

    @pytest.mark.parametrize(
        "list_item",
        json.loads(MODEL_SIMPLE.joinpath("ListItems.json").read_text()),
        ids=lambda item: item.get("provider"),
    )
    def test_list_item(self, list_item):
        list_items = ListItem(**list_item)
        assert isinstance(list_items, ListItem)
        list_items.model_dump_json()

    @pytest.mark.parametrize(
        "raw_item",
        json.loads(MODEL_SIMPLE.joinpath("RawItems.json").read_text()),
        ids=lambda s: s.get("name"),
    )
    def test_raw_item(self, raw_item):
        """"""
        item = RawItem(**raw_item)
        assert item.name == raw_item.get("name")
        item.model_dump_json()

    @pytest.mark.parametrize(
        "hash_info", json.loads(MODEL_SIMPLE.joinpath("HashInfos.json").read_text())
    )
    def test_hash_info(self, hash_info):
        h = HashInfo(**hash_info)
        h.model_dump_json()

    @pytest.mark.parametrize(
        "me", json.loads(MODEL_SIMPLE.joinpath("Me.json").read_text())
    )
    def test_me(self, me):
        Me(**me)

    @pytest.mark.parametrize(
        "storage", json.loads(MODEL_SIMPLE.joinpath("Storages.json").read_text())
    )
    def test_storage(self, storage):
        """"""
        _s = Storage(**storage)
        _s.model_dump_json()

    @pytest.mark.parametrize(
        "url, model, resp",
        json.loads(MODEL_SIMPLE.joinpath("Resps.json").read_text()),
        ids=lambda x: x if isinstance(x, str) else "",
    )
    def test_resp(self, url, model, resp):
        _resp = Resp(**resp)
        if _resp.code == 200:
            assert isinstance(
                _resp.data[0] if isinstance(_resp.data, list) else _resp.data,
                getattr(models, model),
            ), "Model 验证失败~"
